import copy
import gc

import pytest
import torch
import torch.nn as nn

from agilerl.modules.custom_components import NoisyLinear
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import assert_state_dicts_equal, unpack_network


class TwoArgCNN(nn.Module):
    def __init__(self):
        super().__init__()

        # Define the convolutional layers
        self.conv1 = nn.Conv3d(
            in_channels=4, out_channels=16, kernel_size=(2, 3, 3), stride=(4, 4, 4)
        )  # W: 160, H: 210
        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=(1, 3, 3), stride=(2, 2, 2)
        )  # W:

        # Define the max-pooling layers
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)

        # Define fully connected layers
        self.fc1 = nn.Linear(15202, 256)
        self.fc2 = nn.Linear(256, 2)

        # Define activation function
        self.relu = nn.ReLU()

        # Define softmax for classification
        self.softmax = nn.Softmax(dim=1)
        self.tanh = nn.Tanh()

    def forward(self, x, xc):
        # Forward pass through convolutional layers
        x = self.relu(self.conv1(x))
        x = self.relu(self.conv2(x))

        # Flatten the output for the fully connected layers
        x = x.view(x.size(0), -1)
        x = torch.cat([x, xc], dim=1)
        # Forward pass through fully connected layers
        x = self.tanh(self.fc1(x))
        x = self.fc2(x)

        # Apply softmax for classification
        x = self.softmax(x)

        return x


@pytest.fixture
def simple_mlp():
    network = nn.Sequential(
        nn.Linear(10, 20),
        nn.ReLU(),
        nn.Linear(20, 10),
        nn.ReLU(),
        nn.Linear(10, 1),
        nn.Tanh(),
    )
    yield network
    del network
    gc.collect()
    torch.cuda.empty_cache()


@pytest.fixture
def simple_mlp_2():
    network = nn.Sequential(
        nn.Linear(10, 128), nn.ReLU(), nn.Linear(128, 128), nn.ReLU(), nn.Linear(128, 1)
    )
    yield network
    del network
    gc.collect()
    torch.cuda.empty_cache()


@pytest.fixture
def mlp_norm_layers():
    network = nn.Sequential(
        nn.Linear(10, 20),
        nn.LayerNorm(20),
        nn.ReLU(),
        nn.Linear(20, 10),
        nn.LayerNorm(10),
        nn.ReLU(),
        nn.Linear(10, 1),
        nn.Sigmoid(),
    )
    yield network
    del network
    gc.collect()
    torch.cuda.empty_cache()


@pytest.fixture
def simple_cnn():
    network = nn.Sequential(
        nn.Conv2d(
            3, 16, kernel_size=3, stride=1, padding=1
        ),  # Input channels: 3 (for RGB images), Output channels: 16
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Conv2d(
            16, 32, kernel_size=3, stride=1, padding=1
        ),  # Input channels: 16, Output channels: 32
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Flatten(),  # Flatten the 2D feature map to a 1D vector
        nn.Linear(32 * 16 * 16, 128),  # Fully connected layer with 128 output features
        nn.ReLU(),
        nn.Linear(128, 1),  # Output layer with num_classes output features
    )
    yield network
    del network
    gc.collect()
    torch.cuda.empty_cache()


@pytest.fixture
def cnn_norm_layers():
    network = nn.Sequential(
        nn.Conv2d(
            3, 16, kernel_size=3, stride=1, padding=1
        ),  # Input channels: 3 (for RGB images), Output channels: 16
        nn.BatchNorm2d(16),
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Conv2d(
            16, 32, kernel_size=3, stride=1, padding=1
        ),  # Input channels: 16, Output channels: 32
        nn.BatchNorm2d(32),
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Flatten(),  # Flatten the 2D feature map to a 1D vector
        nn.Linear(32 * 16 * 16, 128),  # Fully connected layer with 128 output features
        nn.ReLU(),
        nn.Linear(128, 1),  # Output layer with num_classes output features
    )
    yield network
    del network
    gc.collect()
    torch.cuda.empty_cache()


@pytest.fixture
def two_arg_cnn():
    network = TwoArgCNN()
    yield network
    del network
    gc.collect()
    torch.cuda.empty_cache()


@pytest.fixture
def device():
    return "cuda" if torch.cuda.is_available() else "cpu"


######### Test instantiation #########
# The class can be instantiated with all the required parameters and no errors occur.
@pytest.mark.parametrize(
    "network, input_tensor",
    [("simple_mlp", torch.randn(1, 10)), ("simple_cnn", torch.randn(1, 3, 64, 64))],
)
def test_instantiation_with_required_parameters(network, input_tensor, request):
    network = request.getfixturevalue(network)
    evolvable_network = MakeEvolvable(network, input_tensor)
    assert isinstance(evolvable_network, MakeEvolvable)
    assert str(unpack_network(evolvable_network)) == str(unpack_network(network))


# The class can be instantiated with minimal parameters and default values are assigned correctly.
def test_instantiation_with_minimal_parameters():
    network = nn.Sequential(
        nn.Linear(10, 20), nn.Linear(20, 20), nn.ReLU(), nn.Linear(20, 1), nn.ReLU()
    )
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(network, input_tensor)
    assert isinstance(evolvable_network, MakeEvolvable)


def test_instantiation_with_rainbow():
    network = nn.Sequential(nn.Linear(3, 8), nn.ReLU(), nn.Linear(8, 2))
    input_tensor = torch.randn(1, 3)
    support = torch.linspace(-200, 200, 51)
    evolvable_network = MakeEvolvable(
        network, input_tensor, support=support, rainbow=True
    )
    assert isinstance(evolvable_network, MakeEvolvable)
    assert (
        str(evolvable_network)
        == """MakeEvolvable(
  (feature_net): Sequential(
    (feature_linear_layer_0): Linear(in_features=3, out_features=128, bias=True)
    (feature_activation_0): ReLU()
  )
  (value_net): Sequential(
    (value_linear_layer_0): NoisyLinear(in_features=128, out_features=8)
    (value_activation_0): ReLU()
    (value_linear_layer_output): NoisyLinear(in_features=8, out_features=51)
  )
  (advantage_net): Sequential(
    (advantage_linear_layer_0): NoisyLinear(in_features=128, out_features=8)
    (advantage_activation_0): ReLU()
    (advantage_linear_layer_output): NoisyLinear(in_features=8, out_features=102)
  )
)"""
    )
    del network, evolvable_network


######### Test forward #########
@pytest.mark.parametrize(
    "network, input_tensor, secondary_input_tensor, expected_result",
    [
        ("simple_mlp", torch.randn(1, 10), None, (1, 1)),
        ("simple_cnn", torch.randn(1, 3, 64, 64), None, (1, 1)),
        ("two_arg_cnn", torch.randn(1, 4, 2, 210, 160), torch.randn(1, 2), (1, 2)),
    ],
)
def test_forward_method(
    network, input_tensor, secondary_input_tensor, expected_result, request, device
):
    network = request.getfixturevalue(network)
    if secondary_input_tensor is None:
        evolvable_network = MakeEvolvable(network, input_tensor, device=device)
        with torch.no_grad():
            actual_output = evolvable_network.forward(input_tensor)
    else:
        evolvable_network = MakeEvolvable(
            network,
            input_tensor,
            secondary_input_tensor,
            device=device,
        )
        with torch.no_grad():
            input_tensor = input_tensor.to(dtype=torch.float16)
            actual_output = evolvable_network.forward(
                input_tensor, secondary_input_tensor
            )
    output_shape = actual_output.shape
    assert output_shape == expected_result
    del network, evolvable_network


@pytest.mark.parametrize(
    "network, input_tensor, secondary_input_tensor, expected_result",
    [
        ("simple_mlp", torch.randn(1, 10), None, (1, 1)),
        ("simple_cnn", torch.randn(1, 3, 64, 64), None, (1, 1)),
        ("two_arg_cnn", torch.randn(1, 4, 2, 210, 160), torch.randn(1, 2), (1, 2)),
    ],
)
def test_forward_method_rainbow(
    network, input_tensor, secondary_input_tensor, expected_result, request, device
):
    network = request.getfixturevalue(network)
    support = torch.linspace(-200, 200, 51).to(device)
    if secondary_input_tensor is None:
        evolvable_network = MakeEvolvable(
            network, input_tensor, support=support, rainbow=True, device=device
        )
        with torch.no_grad():
            actual_output = evolvable_network.forward(input_tensor)
    else:
        evolvable_network = MakeEvolvable(
            network,
            input_tensor,
            secondary_input_tensor,
            support=support,
            rainbow=True,
            device=device,
        )
        with torch.no_grad():
            input_tensor = input_tensor.to(dtype=torch.float16)
            actual_output = evolvable_network.forward(
                input_tensor, secondary_input_tensor
            )
    output_shape = actual_output.shape
    assert output_shape == expected_result
    del evolvable_network


# The forward() method can handle different types of input tensors (e.g., numpy array, torch tensor).
def test_forward_method_with_different_input_types(simple_mlp):
    input_tensor = torch.randn(1, 10)
    numpy_array = input_tensor.numpy()
    evolvable_network = MakeEvolvable(simple_mlp, input_tensor)
    with torch.no_grad():
        output1 = evolvable_network.forward(input_tensor)
        output2 = evolvable_network.forward(numpy_array)
    assert isinstance(output1, torch.Tensor)
    assert isinstance(output2, torch.Tensor)
    del evolvable_network


# The forward() method can handle different types of normalization layers (e.g., BatchNorm2d, InstanceNorm3d).
def test_forward_with_different_normalization_layers():
    network = nn.Sequential(
        nn.Linear(10, 20),
        nn.LayerNorm(20),
        nn.ReLU(),
        nn.Linear(20, 10),
        nn.ReLU(),
        nn.Linear(10, 1),
    )
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(network, input_tensor)
    with torch.no_grad():
        output = evolvable_network.forward(input_tensor)
    assert isinstance(output, torch.Tensor)
    assert str(unpack_network(evolvable_network)) == str(unpack_network(network))
    del network, evolvable_network


def test_reset_noise():
    network = nn.Sequential(nn.Linear(3, 8), nn.ReLU(), nn.Linear(8, 2))
    input_tensor = torch.randn(1, 3)
    support = torch.linspace(-200, 200, 51)
    evolvable_mlp = MakeEvolvable(network, input_tensor, support=support, rainbow=True)
    evolvable_mlp.reset_noise()
    assert isinstance(evolvable_mlp.value_net[0], NoisyLinear)
    assert isinstance(evolvable_mlp.advantage_net[0], NoisyLinear)
    del network, evolvable_mlp


######### Test detect architecture function #########


# Detects architecture of a neural network with convolutional layers and without normalization layers
def test_detect_architecture_mlp_simple(device):
    net = nn.Sequential(
        nn.Linear(4, 16), nn.ReLU(), nn.Linear(16, 16), nn.ReLU(), nn.Linear(16, 1)
    )
    evolvable_net = MakeEvolvable(
        net, torch.randn(1, 4), device=device, output_vanish=True
    )
    assert evolvable_net.mlp_layer_info == {"activation_layers": {0: "ReLU", 1: "ReLU"}}
    assert str(unpack_network(net)) == str(unpack_network(evolvable_net))
    del net, evolvable_net


def test_detect_architecture_medium(device):
    net = nn.Sequential(
        nn.Linear(4, 16), nn.Linear(16, 16), nn.ReLU(), nn.Linear(16, 1), nn.Tanh()
    )
    evolvable_net = MakeEvolvable(net, torch.randn(1, 4), device=device)
    assert evolvable_net.mlp_layer_info == {"activation_layers": {1: "ReLU", 2: "Tanh"}}
    assert str(unpack_network(net)) == str(unpack_network(evolvable_net))
    del net, evolvable_net


def test_detect_architecture_complex(device):
    net = nn.Sequential(
        nn.Linear(4, 16),
        nn.LayerNorm(16),
        nn.Linear(16, 16),
        nn.ReLU(),
        nn.Linear(16, 1),
        nn.Tanh(),
    )
    evolvable_net = MakeEvolvable(net, torch.randn(1, 4), device=device)
    assert evolvable_net.mlp_layer_info == {
        "activation_layers": {1: "ReLU", 2: "Tanh"},
        "norm_layers": {0: "LayerNorm"},
    }, evolvable_net.mlp_layer_info
    assert str(unpack_network(net)) == str(unpack_network(evolvable_net))
    del net, evolvable_net


# Test if network after detect arch has the same arch as original network
@pytest.mark.parametrize(
    "network, input_tensor",
    [
        ("simple_mlp", torch.randn(1, 10)),
        ("simple_cnn", torch.randn(1, 3, 64, 64)),
        ("mlp_norm_layers", torch.randn(1, 10)),
        ("cnn_norm_layers", torch.randn(1, 3, 64, 64)),
    ],
)
def test_detect_architecture_networks_the_same(network, input_tensor, device, request):
    network = request.getfixturevalue(network)
    evolvable_network = MakeEvolvable(network, input_tensor, device=device)
    assert str(unpack_network(network)) == str(unpack_network(evolvable_network))
    del network, evolvable_network


def test_detect_architecure_exception(device):
    net = nn.Sequential(
        nn.Linear(4, 16),
        nn.Linear(16, 16),
        nn.ReLU(),
        nn.Linear(16, 1),
        nn.LogSoftmax(dim=-1),
    )
    with pytest.raises(Exception):
        MakeEvolvable(net, torch.randn(1, 4), device=device)
    del net


def test_detect_architecture_empty_hidden_size(device):
    net = nn.Sequential(nn.Linear(2, 10))
    with pytest.raises(TypeError):
        MakeEvolvable(net, torch.randn(1, 2), device=device)
    del net


def test_detect_architecure_different_acitvations(device):
    net = nn.Sequential(
        nn.Linear(2, 16), nn.ReLU(), nn.Linear(16, 16), nn.Tanh(), nn.Linear(16, 4)
    )
    with pytest.raises(TypeError):
        MakeEvolvable(net, torch.randn(1, 2), device=device)
    del net


######### Test add_mlp_layer #########


def test_add_mlp_layer_simple(simple_mlp, device):
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(
        simple_mlp, input_tensor, init_layers=True, device=device
    )
    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())
    initial_num_layers = len(evolvable_network.hidden_size)
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {0: "ReLU", 1: "ReLU", 2: "Tanh"}
    }, evolvable_network.mlp_layer_info
    evolvable_network.add_mlp_layer()
    new_feature_net = evolvable_network.feature_net
    assert len(evolvable_network.hidden_size) == initial_num_layers + 1
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {0: "ReLU", 1: "ReLU", 2: "ReLU", 3: "Tanh"}
    }, evolvable_network.mlp_layer_info
    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])
    del evolvable_network


def test_add_mlp_layer_medium(device):
    network = nn.Sequential(
        nn.Linear(4, 16), nn.Linear(16, 16), nn.ReLU(), nn.Linear(16, 1), nn.Tanh()
    )
    evolvable_network = MakeEvolvable(network, torch.randn(1, 4), device=device)
    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())
    initial_num_layers = len(evolvable_network.hidden_size)
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "ReLU", 2: "Tanh"}
    }, evolvable_network.mlp_layer_info
    evolvable_network.add_mlp_layer()
    new_feature_net = evolvable_network.feature_net
    assert len(evolvable_network.hidden_size) == initial_num_layers + 1
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "ReLU", 2: "ReLU", 3: "Tanh"}
    }, evolvable_network.mlp_layer_info
    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])
    del network, evolvable_network


def test_add_mlp_layer_complex(device):
    net = nn.Sequential(
        nn.Linear(4, 16),
        nn.LayerNorm(16),
        nn.Linear(16, 16),
        nn.ReLU(),
        nn.Linear(16, 1),
        nn.Tanh(),
    )
    evolvable_network = MakeEvolvable(net, torch.randn(1, 4), device=device)
    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())
    initial_num_layers = len(evolvable_network.hidden_size)
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "ReLU", 2: "Tanh"},
        "norm_layers": {0: "LayerNorm"},
    }, evolvable_network.mlp_layer_info
    evolvable_network.add_mlp_layer()
    new_feature_net = evolvable_network.feature_net
    assert len(evolvable_network.hidden_size) == initial_num_layers + 1
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "ReLU", 2: "ReLU", 3: "Tanh"},
        "norm_layers": {0: "LayerNorm"},
    }, evolvable_network.mlp_layer_info
    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])
    del net, evolvable_network


def test_add_mlp_layer_else_statement(device):
    network = nn.Sequential(
        nn.Linear(4, 16), nn.Linear(16, 16), nn.ReLU(), nn.Linear(16, 1), nn.Tanh()
    )
    evolvable_network = MakeEvolvable(
        network, torch.randn(1, 4), device=device, max_hidden_layers=2
    )
    initial_num_layers = len(evolvable_network.hidden_size)
    evolvable_network.add_mlp_layer()
    assert initial_num_layers == len(evolvable_network.hidden_size)
    del network, evolvable_network


######### Test remove_mlp_layer #########
def test_remove_mlp_layer_simple(simple_mlp_2, device):
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(simple_mlp_2, input_tensor, device=device)
    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())
    initial_num_layers = len(evolvable_network.hidden_size)
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {0: "ReLU", 1: "ReLU"}
    }
    evolvable_network.remove_mlp_layer()
    new_feature_net = evolvable_network.feature_net
    assert len(evolvable_network.hidden_size) == initial_num_layers - 1
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {0: "ReLU"}
    }, evolvable_network.mlp_layer_info
    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            torch.testing.assert_close(param, feature_net_dict[key])
    del evolvable_network


def test_remove_mlp_layer_medium(device):
    network = nn.Sequential(
        nn.Linear(4, 16), nn.Linear(16, 16), nn.ReLU(), nn.Linear(16, 1), nn.Tanh()
    )
    evolvable_network = MakeEvolvable(network, torch.randn(1, 4), device=device)
    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())
    initial_num_layers = len(evolvable_network.hidden_size)
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "ReLU", 2: "Tanh"}
    }, evolvable_network.mlp_layer_info
    evolvable_network.remove_mlp_layer()
    new_feature_net = evolvable_network.feature_net
    assert len(evolvable_network.hidden_size) == initial_num_layers - 1
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "Tanh"}
    }, evolvable_network.mlp_layer_info
    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])
    del network, evolvable_network


def test_remove_mlp_layer_complex(device):
    net = nn.Sequential(
        nn.Linear(4, 16),
        nn.LayerNorm(16),
        nn.Linear(16, 16),
        nn.LayerNorm(16),
        nn.ReLU(),
        nn.Linear(16, 1),
        nn.Tanh(),
    )
    evolvable_network = MakeEvolvable(net, torch.randn(1, 4), device=device)
    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())
    initial_num_layers = len(evolvable_network.hidden_size)
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "ReLU", 2: "Tanh"},
        "norm_layers": {0: "LayerNorm", 1: "LayerNorm"},
    }, evolvable_network.mlp_layer_info
    evolvable_network.remove_mlp_layer()
    new_feature_net = evolvable_network.feature_net
    assert len(evolvable_network.hidden_size) == initial_num_layers - 1
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {1: "Tanh"},
        "norm_layers": {0: "LayerNorm"},
    }, evolvable_network.mlp_layer_info
    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])
    del net, evolvable_network


def test_remove_mlp_layer_else_statement(device):
    network = nn.Sequential(
        nn.Linear(4, 16), nn.Linear(16, 16), nn.ReLU(), nn.Linear(16, 1), nn.Tanh()
    )
    evolvable_network = MakeEvolvable(
        network, torch.randn(1, 4), device=device, min_hidden_layers=2
    )
    initial_num_layers = len(evolvable_network.hidden_size)
    evolvable_network.remove_mlp_layer()
    assert initial_num_layers == len(evolvable_network.hidden_size)
    del network, evolvable_network


def test_remove_mlp_layer_no_output_activation(device):
    net = nn.Sequential(
        nn.Linear(4, 16), nn.ReLU(), nn.Linear(16, 16), nn.Linear(16, 1), nn.Tanh()
    )
    evolvable_network = MakeEvolvable(net, torch.randn(1, 4), device=device)
    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())
    initial_num_layers = len(evolvable_network.hidden_size)
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {0: "ReLU", 2: "Tanh"}
    }, evolvable_network.mlp_layer_info
    evolvable_network.remove_mlp_layer()
    new_feature_net = evolvable_network.feature_net
    assert len(evolvable_network.hidden_size) == initial_num_layers - 1
    assert evolvable_network.mlp_layer_info == {
        "activation_layers": {0: "ReLU", 1: "Tanh"}
    }, evolvable_network.mlp_layer_info
    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])
    del net, evolvable_network


######### Test add_mlp_node #########
def test_add_mlp_node_fixed(simple_mlp, device):
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(simple_mlp, input_tensor, device=device)

    # Test adding a new node to a specific layer
    hidden_layer = 1
    numb_new_nodes = 8
    result = evolvable_network.add_mlp_node(hidden_layer, numb_new_nodes)

    # Check if the hidden layer and number of new nodes are updated correctly
    assert evolvable_network.hidden_size[hidden_layer] == 18
    assert result["hidden_layer"] == hidden_layer
    assert result["numb_new_nodes"] == numb_new_nodes
    del evolvable_network


######### Test remove_mlp_node #########
def test_remove_mlp_node(simple_mlp_2, device):
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(simple_mlp_2, input_tensor, device=device)

    # Check the initial number of nodes in the hidden layers
    assert len(evolvable_network.hidden_size) == 2

    # Remove a node from the second hidden layer
    evolvable_network.remove_mlp_node(hidden_layer=1, numb_new_nodes=10)

    # Check that the number of nodes in the second hidden layer has decreased by 10
    assert evolvable_network.hidden_size[1] == 118

    # Remove a node from the first hidden layer
    evolvable_network.remove_mlp_node(hidden_layer=0, numb_new_nodes=5)

    # Check that the number of nodes in the first hidden layer has decreased by 5
    assert evolvable_network.hidden_size[0] == 123

    del evolvable_network


def test_remove_mlp_node_no_arg(device):
    mlp = nn.Sequential(
        nn.Linear(10, 256),
        nn.ReLU(),
        nn.Linear(256, 128),
        nn.ReLU(),
        nn.Linear(128, 10),
    )
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(mlp, input_tensor, device=device, min_mlp_nodes=2)
    original_hidden_size = copy.deepcopy(evolvable_network.hidden_size)
    # Test adding a new node to a specific layer
    hidden_layer = None
    numb_new_nodes = None
    result = evolvable_network.remove_mlp_node(hidden_layer, numb_new_nodes)
    hidden_layer = result["hidden_layer"]
    numb_new_nodes = result["numb_new_nodes"]

    # Check if the hidden layer and number of new nodes are updated correctly
    assert (
        evolvable_network.hidden_size[hidden_layer]
        == original_hidden_size[hidden_layer] - numb_new_nodes
    )
    del mlp, evolvable_network


######### Test add_cnn_layer #########
def test_make_evo_add_cnn_layer(simple_cnn, device):
    input_tensor = torch.randn(1, 3, 64, 64)
    evolvable_network = MakeEvolvable(simple_cnn, input_tensor, device=device)

    # Check the initial number of layers
    assert len(evolvable_network.channel_size) == 2

    # Add a new CNN layer
    evolvable_network.add_cnn_layer()

    # Check if a new layer has been added
    assert len(evolvable_network.channel_size) == 3
    assert evolvable_network.cnn_layer_info == {
        "activation_layers": {0: "ReLU", 1: "ReLU", 2: "ReLU"},
        "conv_layer_type": "Conv2d",
        "pooling_layers": {
            0: {"name": "MaxPool2d", "kernel": 2, "stride": 2, "padding": 0},
            1: {"name": "MaxPool2d", "kernel": 2, "stride": 2, "padding": 0},
        },
    }, evolvable_network.cnn_layer_info
    del evolvable_network


def test_make_evo_add_cnn_layer_multi(two_arg_cnn, device):
    evolvable_cnn = MakeEvolvable(
        two_arg_cnn,
        torch.randn(1, 4, 2, 210, 160),
        torch.randn(1, 2),
        device=device,
    )
    original_channels = copy.deepcopy(evolvable_cnn.channel_size)
    evolvable_cnn.add_cnn_layer()
    assert len(original_channels) + 1 == len(evolvable_cnn.channel_size)
    del evolvable_cnn


def test_make_evo_add_cnn_layer_no_activation(device):
    cnn = nn.Sequential(
        nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1),
        nn.Conv2d(16, 32, kernel_size=3, stride=1, padding=1),
        nn.Flatten(),
        nn.Linear(32768, 128),  # Fully connected layer with 128 output features
        nn.ReLU(),
        nn.Linear(128, 4),
    )
    evolvable_cnn = MakeEvolvable(cnn, torch.randn(1, 3, 32, 32), device=device)
    original_channels = copy.deepcopy(evolvable_cnn.channel_size)
    evolvable_cnn.add_cnn_layer()
    assert len(original_channels) + 1 == len(evolvable_cnn.channel_size)
    del cnn, evolvable_cnn


def test_make_evo_add_cnn_layer_else_statement(simple_cnn, device):
    evolvable_cnn = MakeEvolvable(
        simple_cnn, torch.randn(1, 3, 64, 64), device=device, max_cnn_hidden_layers=2
    )
    original_channels = copy.deepcopy(evolvable_cnn.channel_size)
    evolvable_cnn.add_cnn_layer()
    assert len(original_channels) == len(evolvable_cnn.channel_size)
    del evolvable_cnn


######### Test remove_cnn_layer #########
def test_remove_cnn_layer(simple_cnn, device):
    input_tensor = torch.randn(1, 3, 64, 64)
    evolvable_network = MakeEvolvable(simple_cnn, input_tensor, device=device)

    # Check the initial number of layers
    assert len(evolvable_network.channel_size) == 2

    # Remove a CNN layer
    evolvable_network.remove_cnn_layer()

    # Check if a layer has been removed
    assert len(evolvable_network.channel_size) == 1
    assert evolvable_network.cnn_layer_info == {
        "activation_layers": {0: "ReLU"},
        "conv_layer_type": "Conv2d",
        "pooling_layers": {
            0: {"name": "MaxPool2d", "kernel": 2, "stride": 2, "padding": 0}
        },
    }, evolvable_network.cnn_layer_info
    del evolvable_network


def test_remove_cnn_layer_no_activation(device):
    cnn = nn.Sequential(
        nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1),
        nn.Conv2d(16, 32, kernel_size=3, stride=1, padding=1),
        nn.BatchNorm2d(32),
        nn.Flatten(),
        nn.Linear(32768, 128),  # Fully connected layer with 128 output features
        nn.ReLU(),
        nn.Linear(128, 4),
    )
    evolvable_cnn = MakeEvolvable(cnn, torch.randn(1, 3, 32, 32), device=device)
    original_channels = copy.deepcopy(evolvable_cnn.channel_size)
    evolvable_cnn.remove_cnn_layer()
    assert len(original_channels) - 1 == len(evolvable_cnn.channel_size)
    del cnn, evolvable_cnn


def test_remove_cnn_layer_else_statement(simple_cnn, device):
    evolvable_cnn = MakeEvolvable(
        simple_cnn, torch.randn(1, 3, 64, 64), device=device, min_cnn_hidden_layers=2
    )
    original_channels = copy.deepcopy(evolvable_cnn.channel_size)
    evolvable_cnn.remove_cnn_layer()
    assert len(original_channels) == len(evolvable_cnn.channel_size)
    del evolvable_cnn


######### Test add_cnn_channel #########
def test_make_evo_add_cnn_channel(simple_cnn, device):
    input_tensor = torch.randn(1, 3, 64, 64)
    evolvable_network = MakeEvolvable(simple_cnn, input_tensor, device=device)
    numb_new_channels = 16
    layer = 1
    original_channel_size = evolvable_network.channel_size[layer]
    evolvable_network.add_cnn_channel(
        hidden_layer=layer, numb_new_channels=numb_new_channels
    )
    assert evolvable_network.channel_size[layer] == original_channel_size + 16
    del evolvable_network


######### Test remove_cnn_channel #########
def test_remove_cnn_channel(device):
    input_tensor = torch.randn(1, 3, 64, 64)
    cnn = nn.Sequential(
        nn.Conv2d(3, 256, 3, 2),
        nn.ReLU(),
        nn.Conv2d(256, 128, 3, 2),
        nn.ReLU(),
        nn.Flatten(),
        nn.Linear(28800, 128),
        nn.ReLU(),
        nn.Linear(128, 10),
    )
    evolvable_network = MakeEvolvable(cnn, input_tensor, device=device)
    numb_new_channels = None
    layer = None
    original_channel_size = copy.deepcopy(evolvable_network.channel_size)
    evolvable_network.min_channel_size = 16
    result = evolvable_network.remove_cnn_channel(
        hidden_layer=layer, numb_new_channels=numb_new_channels
    )
    numb_new_channels = result["numb_new_channels"]
    hidden_layer = result["hidden_layer"]
    assert (
        evolvable_network.channel_size[hidden_layer]
        == original_channel_size[hidden_layer] - numb_new_channels
    )
    del cnn, evolvable_network


def test_remove_cnn_channel_specified_hidden_layer(device):
    input_tensor = torch.randn(1, 3, 64, 64)
    cnn = nn.Sequential(
        nn.Conv2d(3, 256, 3, 2),
        nn.ReLU(),
        nn.Conv2d(256, 128, 3, 2),
        nn.ReLU(),
        nn.Flatten(),
        nn.Linear(28800, 128),
        nn.ReLU(),
        nn.Linear(128, 10),
    )
    evolvable_network = MakeEvolvable(cnn, input_tensor, device=device)
    numb_new_channels = None
    layer = 0
    original_channel_size = copy.deepcopy(evolvable_network.channel_size)
    evolvable_network.min_channel_size = 16
    result = evolvable_network.remove_cnn_channel(
        hidden_layer=layer, numb_new_channels=numb_new_channels
    )
    numb_new_channels = result["numb_new_channels"]
    assert (
        evolvable_network.channel_size[layer]
        == original_channel_size[layer] - numb_new_channels
    )
    del cnn, evolvable_network


######### Test change_cnn_kernel #########
def test_change_cnn_kernel(simple_cnn, device):
    input_tensor = torch.randn(1, 3, 64, 64)
    evolvable_network = MakeEvolvable(simple_cnn, input_tensor, device=device)

    # Check initial kernel sizes
    assert evolvable_network.kernel_size == [(3, 3), (3, 3)]

    # Change kernel size
    evolvable_network.change_cnn_kernel()

    while evolvable_network.kernel_size == [(3, 3), (3, 3)]:
        evolvable_network.change_cnn_kernel()

    # Check if kernel size has changed
    assert evolvable_network.kernel_size != [
        (3, 3),
        (3, 3),
    ], evolvable_network.kernel_size
    del evolvable_network


def test_change_cnn_kernel_else_statement(device):
    network = nn.Sequential(
        nn.Conv2d(3, 16, kernel_size=(3, 3), stride=1, padding=1),
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Flatten(),
        nn.Linear(16384, 128),
        nn.ReLU(),
        nn.Linear(128, 1),
    )
    input_tensor = torch.randn(1, 3, 64, 64)
    evolvable_network = MakeEvolvable(network, input_tensor, device=device)

    # Change kernel size
    evolvable_network.change_cnn_kernel()
    # Check if kernel size has changed
    assert len(evolvable_network.kernel_size) == 2
    assert len(evolvable_network.kernel_size[-1]) == 2
    del network, evolvable_network


def test_change_kernel_multi_single_arg(device):
    network = nn.Sequential(
        nn.Conv3d(3, 16, kernel_size=(1, 3, 3), stride=(1, 3, 3), padding=1),
        nn.ReLU(),
        nn.Conv3d(16, 32, kernel_size=(1, 3, 3), stride=(1, 3, 3), padding=1),
        nn.ReLU(),
        nn.Flatten(),
        nn.Linear(69120, 128),
        nn.ReLU(),
        nn.Linear(128, 1),
    )
    evolvable_cnn = MakeEvolvable(
        network, torch.randn(1, 3, 1, 210, 160), device=device
    )
    while evolvable_cnn.kernel_size == [(1, 3, 3), (1, 3, 3)]:
        evolvable_cnn.change_cnn_kernel()

    assert evolvable_cnn.kernel_size != [
        (1, 3, 3),
        (1, 3, 3),
    ], evolvable_cnn.kernel_size
    del network, evolvable_cnn


def test_change_kernel_multi_single_arg_else_statement(device):
    network = nn.Sequential(
        nn.Conv3d(3, 16, kernel_size=(1, 3, 3), stride=(1, 3, 3), padding=1),
        nn.ReLU(),
        nn.Flatten(),
        nn.Linear(181440, 128),
        nn.ReLU(),
        nn.Linear(128, 1),
    )
    evolvable_cnn = MakeEvolvable(
        network, torch.randn(1, 3, 1, 210, 160), device=device
    )
    evolvable_cnn.change_cnn_kernel()
    assert len(evolvable_cnn.kernel_size) == 2
    assert len(evolvable_cnn.kernel_size[-1]) == 3
    del network, evolvable_cnn


def test_change_kernel_multi_two_arg(two_arg_cnn, device):
    evolvable_cnn = MakeEvolvable(
        two_arg_cnn,
        torch.randn(1, 4, 2, 210, 160),
        torch.randn(1, 2),
        device=device,
    )
    while evolvable_cnn.kernel_size == [(2, 3, 3), (1, 3, 3)]:
        evolvable_cnn.change_cnn_kernel()

    assert evolvable_cnn.kernel_size != [
        (2, 3, 3),
        (1, 3, 3),
    ], evolvable_cnn.kernel_size
    del evolvable_cnn


######### Test recreate_nets #########
def test_recreate_nets_parameters_preserved(simple_mlp, device):
    input_tensor = torch.randn(1, 10)
    evolvable_network = MakeEvolvable(simple_mlp, input_tensor, device=device)

    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())

    # Modify the architecture
    evolvable_network.hidden_size += [evolvable_network.hidden_size[-1]]

    evolvable_network.recreate_network()
    new_feature_net = evolvable_network.feature_net

    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])
    del evolvable_network


def test_recreate_nets_parameters_preserved_rainbow(simple_mlp, device):
    input_tensor = torch.randn(1, 10)
    support = torch.linspace(-200, 200, 51).to(device)
    evolvable_network = MakeEvolvable(
        simple_mlp, input_tensor, support=support, rainbow=True, device=device
    )

    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())

    value_net = evolvable_network.value_net
    value_net_dict = dict(value_net.named_parameters())

    advantage_net = evolvable_network.advantage_net
    advantage_net_dict = dict(advantage_net.named_parameters())

    # Modify the architecture
    evolvable_network.hidden_size += [evolvable_network.hidden_size[-1]]

    evolvable_network.recreate_network()
    new_feature_net = evolvable_network.feature_net
    new_value_net = evolvable_network.value_net
    new_advantage_net = evolvable_network.advantage_net

    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            assert torch.equal(param, feature_net_dict[key])

    for key, param in new_value_net.named_parameters():
        if key in value_net_dict.keys():
            assert torch.equal(param, value_net_dict[key])

    for key, param in new_advantage_net.named_parameters():
        if key in advantage_net_dict.keys():
            assert torch.equal(param, advantage_net_dict[key])
    del evolvable_network


def test_recreate_nets_parameters_shrink_preserved(device):
    network = nn.Sequential(
        nn.Linear(4, 32), nn.ReLU(), nn.Linear(32, 32), nn.ReLU(), nn.Linear(32, 2)
    )

    input_tensor = torch.randn(1, 4)
    evolvable_network = MakeEvolvable(network, input_tensor, device=device)

    feature_net = evolvable_network.feature_net
    feature_net_dict = dict(feature_net.named_parameters())

    # Modify the architecture
    evolvable_network.hidden_size = evolvable_network.hidden_size[:-1]
    evolvable_network.recreate_network(shrink_params=True)
    new_feature_net = evolvable_network.feature_net

    for key, param in new_feature_net.named_parameters():
        if key in feature_net_dict.keys():
            torch.testing.assert_close(param, feature_net_dict[key])
    del network, evolvable_network


######### Test clone #########


# The clone() method successfully creates a deep copy of the model.
@pytest.mark.parametrize(
    "network, input_tensor, secondary_input_tensor",
    [
        ("simple_mlp", torch.randn(1, 10), None),
        ("simple_cnn", torch.randn(1, 3, 64, 64), None),
        ("two_arg_cnn", torch.randn(1, 4, 2, 210, 160), torch.randn(1, 2)),
    ],
)
def test_clone_method_with_equal_state_dicts(
    network,
    input_tensor,
    secondary_input_tensor,
    request,
    device,
):
    network = request.getfixturevalue(network)
    if secondary_input_tensor is None:
        evolvable_network = MakeEvolvable(network, input_tensor, device=device)
    else:
        evolvable_network = MakeEvolvable(
            network,
            input_tensor,
            secondary_input_tensor,
            device=device,
        )
    clone_network = evolvable_network.clone()
    assert isinstance(clone_network, MakeEvolvable)
    assert_state_dicts_equal(evolvable_network.state_dict(), clone_network.state_dict())
    del evolvable_network, clone_network



from typing import Any

import torch
import torch.nn.functional as F

from agilerl.modules.base import EvolvableModule
from agilerl.modules.mlp import EvolvableMLP
from agilerl.utils.evolvable_networks import create_mlp


class DuelingDistributionalMLP(EvolvableMLP):
    """A multi-layer perceptron network that calculates state-action values through
    the use of separate advantage and value networks. It outputs a distribution of values
    for both of these networks. Used in the Rainbow DQN algorithm.

    :param num_inputs: Number of input features.
    :type num_inputs: int
    :param num_outputs: Number of output features.
    :type num_outputs: int
    :param hidden_size: List of hidden layer sizes.
    :type hidden_size: list[int]
    :param num_atoms: Number of atoms in the distribution.
    :type num_atoms: int
    :param support: Support of the distribution.
    :type support: torch.Tensor
    :param layer_norm: Normalization between layers, defaults to True
    :type layer_norm: bool, optional
    :param output_layernorm: Normalization after the last layer, defaults to False
    :type output_layernorm: bool, optional
    :param output_vanish: Vanish output by multiplying by 0.1, defaults to True
    :type output_vanish: bool, optional
    :param init_layers: Initialise network layers, defaults to True
    :type init_layers: bool, optional
    :param noisy: Use noisy layers, defaults to True
    :type noisy: bool, optional
    :param noise_std: Standard deviation of the noise. Defaults to 0.5.
    :type noise_std: float, optional
    :param activation: Activation layer, defaults to 'ReLU'
    :type activation: str, optional
    :param output_activation: Output activation layer, defaults to None
    :type output_activation: str, optional
    :param min_hidden_layers: Minimum number of hidden layers the network will shrink down to, defaults to 1
    :type min_hidden_layers: int, optional
    :param max_hidden_layers: Maximum number of hidden layers the network will expand to, defaults to 3
    :type max_hidden_layers: int, optional
    :param min_mlp_nodes: Minimum number of nodes a layer can have within the network, defaults to 64
    :type min_mlp_nodes: int, optional
    :param max_mlp_nodes: Maximum number of nodes a layer can have within the network, defaults to 500
    :type max_mlp_nodes: int, optional
    :param new_gelu: Use new GELU activation function, defaults to False
    :type new_gelu: bool, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    """

    def __init__(
        self,
        num_inputs: int,
        num_outputs: int,
        hidden_size: list[int],
        num_atoms: int,
        support: torch.Tensor,
        layer_norm: bool = True,
        output_layernorm: bool = False,
        output_vanish: bool = True,
        init_layers: bool = False,
        noisy: bool = True,
        noise_std: float = 0.5,
        activation: str = "ReLU",
        output_activation: str = None,
        min_hidden_layers: int = 1,
        max_hidden_layers: int = 3,
        min_mlp_nodes: int = 64,
        max_mlp_nodes: int = 500,
        new_gelu: bool = False,
        device: str = "cpu",
    ) -> None:

        super().__init__(
            num_inputs,
            num_atoms,
            hidden_size,
            activation,
            output_activation,
            min_hidden_layers,
            max_hidden_layers,
            min_mlp_nodes,
            max_mlp_nodes,
            layer_norm=layer_norm,
            output_layernorm=output_layernorm,
            output_vanish=output_vanish,
            init_layers=init_layers,
            noisy=noisy,
            noise_std=noise_std,
            new_gelu=new_gelu,
            device=device,
            name="value",
        )

        self.num_atoms = num_atoms
        self.num_actions = num_outputs
        self.support = support

        self.advantage_net = create_mlp(
            input_size=num_inputs,
            output_size=num_outputs * num_atoms,
            hidden_size=self.hidden_size,
            output_vanish=self.output_vanish,
            output_activation=self.output_activation,
            noisy=self.noisy,
            init_layers=self.init_layers,
            layer_norm=self.layer_norm,
            activation=self.activation,
            noise_std=self.noise_std,
            device=self.device,
            new_gelu=self.new_gelu,
            name="advantage",
        )

    @property
    def net_config(self) -> dict[str, Any]:
        net_config = super().net_config.copy()
        net_config.pop("num_atoms")
        net_config.pop("support")
        return net_config

    def forward(
        self, x: torch.Tensor, q: bool = True, log: bool = False
    ) -> torch.Tensor:
        """Forward pass of the network.

        :param obs: Input to the network.
        :type obs: torch.Tensor, dict[str, torch.Tensor], or list[torch.Tensor]
        :param q: Whether to return Q values. Defaults to True.
        :type q: bool
        :param log: Whether to return log probabilities. Defaults to False.
        :type log: bool

        :return: Output of the network.
        :rtype: torch.Tensor
        """
        value: torch.Tensor = self.model(x)
        advantage: torch.Tensor = self.advantage_net(x)

        batch_size = value.size(0)
        value = value.view(batch_size, 1, self.num_atoms)
        advantage = advantage.view(batch_size, self.num_actions, self.num_atoms)

        x = value + advantage - advantage.mean(1, keepdim=True)
        if log:
            x = F.log_softmax(x.view(-1, self.num_atoms), dim=-1)
            return x.view(-1, self.num_actions, self.num_atoms)

        x = F.softmax(x.view(-1, self.num_atoms), dim=-1)
        x = x.view(-1, self.num_actions, self.num_atoms).clamp(min=1e-3)
        if q:
            x = torch.sum(x * self.support, dim=2)

        return x

    def recreate_network(self) -> None:
        """Recreates the network with the same parameters."""
        # Recreate value net with the same parameters
        super().recreate_network()

        advantage_net = create_mlp(
            input_size=self.num_inputs,
            output_size=self.num_actions * self.num_atoms,
            hidden_size=self.hidden_size,
            output_activation=self.output_activation,
            output_vanish=self.output_vanish,
            noisy=self.noisy,
            init_layers=self.init_layers,
            layer_norm=self.layer_norm,
            activation=self.activation,
            noise_std=self.noise_std,
            device=self.device,
            new_gelu=self.new_gelu,
            name="advantage",
        )

        self.advantage_net = EvolvableModule.preserve_parameters(
            self.advantage_net, advantage_net
        )



from unittest.mock import Mock

from agilerl.modules.base import EvolvableModule, EvolvableWrapper, ModuleDict, mutation
from agilerl.protocols import MutationType


def test_register_mutation_fn():
    @mutation(MutationType.NODE)
    def dummy_mutation(self):
        return {"mutation": "dummy"}

    assert dummy_mutation._mutation_type == MutationType.NODE


def test_evolvable_module_initialization():
    class DummyEvolvableModule(EvolvableModule):
        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = DummyEvolvableModule(device="cpu")
    assert module.device == "cpu"
    assert module.get_init_dict() == {"device": "cpu", "random_seed": None}


def test_evolvable_module_get_mutation_methods():
    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_mutation(self):
            return {"mutation": "dummy"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = DummyEvolvableModule(device="cpu")
    mutation_methods = module.get_mutation_methods()
    assert "dummy_mutation" in mutation_methods
    assert mutation_methods["dummy_mutation"]._mutation_type == MutationType.NODE


def test_evolvable_module_clone():
    class DummyEvolvableModule(EvolvableModule):
        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = DummyEvolvableModule(device="cpu")
    clone = module.clone()
    assert clone.device == module.device
    assert clone.get_init_dict() == module.get_init_dict()


def test_evolvable_module_make_unevolvable():
    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_mutation(self):
            return {"mutation": "dummy"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = DummyEvolvableModule(device="cpu")
    module.disable_mutations()
    assert module.mutation_methods == []
    assert module.layer_mutation_methods == []
    assert module.node_mutation_methods == []


def test_evolvable_module_sample_mutation_method():
    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_mutation(self):
            return {"mutation": "dummy"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = DummyEvolvableModule(device="cpu")
    sampled_method = module.sample_mutation_method(new_layer_prob=0.5)
    assert sampled_method == "dummy_mutation"


def test_inherited_evolvable_module_mutation_methods():
    class BaseEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def base_mutation(self):
            return {"mutation": "base"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    class InheritedEvolvableModule(BaseEvolvableModule):
        @mutation(MutationType.LAYER)
        def inherited_mutation(self):
            return {"mutation": "inherited"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = InheritedEvolvableModule(device="cpu")
    mutation_methods = module.get_mutation_methods()
    assert "base_mutation" in mutation_methods
    assert "inherited_mutation" in mutation_methods
    assert mutation_methods["base_mutation"]._mutation_type == MutationType.NODE
    assert mutation_methods["inherited_mutation"]._mutation_type == MutationType.LAYER


def test_evolvable_module_with_evolvable_attributes():
    class AttributeEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def attribute_mutation(self):
            return {"mutation": "attribute"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    class ParentEvolvableModule(EvolvableModule):
        def __init__(self, device="cpu"):
            super().__init__(device)
            self.attribute_module = AttributeEvolvableModule(device=device)

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = ParentEvolvableModule(device="cpu")
    mutation_methods = module.get_mutation_methods()
    assert "attribute_module.attribute_mutation" in mutation_methods
    assert (
        mutation_methods["attribute_module.attribute_mutation"]._mutation_type
        == MutationType.NODE
    )


def test_mutation_calls_recreate_network():
    """Test that the mutation decorator calls recreate_network after mutation."""

    class MockEvolvableModule(EvolvableModule):
        def __init__(self, device="cpu"):
            super().__init__(device)

        @mutation(MutationType.NODE)
        def node_mutation(self):
            return {"mutation": "node"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create a module instance
    module = MockEvolvableModule(device="cpu")

    # Replace recreate_network with a mock
    module.recreate_network = Mock()

    # Call the mutation method, which should trigger recreate_network via the wrapper
    module.node_mutation()

    # Verify recreate_network was called once
    module.recreate_network.assert_called_once_with()


def test_mutation_with_recreate_kwargs():
    """Test that the kwargs are correctly passed to the mutation decorator."""
    # Instead of testing the actual call to recreate_network with shrink_params,
    # test that the mutation decorator correctly sets the _recreate_kwargs attribute

    @mutation(MutationType.NODE, shrink_params=True)
    def test_mutation(self):
        pass

    # Check that the decorator correctly set the _recreate_kwargs attribute
    assert test_mutation._recreate_kwargs == {"shrink_params": True}


def test_mutation_with_args_kwargs():
    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def mutation_with_args(self, arg1, arg2=None):
            return {"arg1": arg1, "arg2": arg2}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    module = DummyEvolvableModule(device="cpu")
    module.recreate_network = Mock()

    result = module.mutation_with_args("test", arg2="value")

    assert result == {"arg1": "test", "arg2": "value"}
    module.recreate_network.assert_called_once()


######### Test ModuleDict #########
def test_module_dict_initialization():
    """Test ModuleDict initialization with evolvable modules."""

    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_node_mutation(self):
            return {"mutation": "node"}

        @mutation(MutationType.LAYER)
        def dummy_layer_mutation(self):
            return {"mutation": "layer"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create a ModuleDict with device parameter first
    module_dict = ModuleDict(device="cpu")

    # Add modules after initialization
    module_dict["module1"] = DummyEvolvableModule(device="cpu")
    module_dict["module2"] = DummyEvolvableModule(device="cpu")

    # Check that the ModuleDict contains the modules
    assert "module1" in module_dict
    assert "module2" in module_dict
    assert isinstance(module_dict["module1"], DummyEvolvableModule)
    assert isinstance(module_dict["module2"], DummyEvolvableModule)


def test_module_dict_mutation_methods():
    """Test that ModuleDict correctly exposes mutation methods from contained modules."""

    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_node_mutation(self):
            return {"mutation": "node"}

        @mutation(MutationType.LAYER)
        def dummy_layer_mutation(self):
            return {"mutation": "layer"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create a ModuleDict
    module_dict = ModuleDict(device="cpu")

    # Add modules after initialization
    module_dict["module1"] = DummyEvolvableModule(device="cpu")
    module_dict["module2"] = DummyEvolvableModule(device="cpu")

    # Check that mutation methods are correctly exposed
    mutation_methods = module_dict.get_mutation_methods()

    # Check that all expected mutation methods are present
    assert "module1.dummy_node_mutation" in mutation_methods
    assert "module1.dummy_layer_mutation" in mutation_methods
    assert "module2.dummy_node_mutation" in mutation_methods
    assert "module2.dummy_layer_mutation" in mutation_methods

    # Check mutation types
    assert (
        mutation_methods["module1.dummy_node_mutation"]._mutation_type
        == MutationType.NODE
    )
    assert (
        mutation_methods["module1.dummy_layer_mutation"]._mutation_type
        == MutationType.LAYER
    )


def test_module_dict_layer_node_methods():
    """Test the layer_mutation_methods and node_mutation_methods properties of ModuleDict."""

    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_node_mutation(self):
            return {"mutation": "node"}

        @mutation(MutationType.LAYER)
        def dummy_layer_mutation(self):
            return {"mutation": "layer"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create a ModuleDict
    module_dict = ModuleDict(device="cpu")

    # Add modules after initialization
    module_dict["module1"] = DummyEvolvableModule(device="cpu")
    module_dict["module2"] = DummyEvolvableModule(device="cpu")

    # Check layer mutation methods
    layer_methods = module_dict.layer_mutation_methods
    assert "module1.dummy_layer_mutation" in layer_methods
    assert "module2.dummy_layer_mutation" in layer_methods
    assert "module1.dummy_node_mutation" not in layer_methods

    # Check node mutation methods
    node_methods = module_dict.node_mutation_methods
    assert "module1.dummy_node_mutation" in node_methods
    assert "module2.dummy_node_mutation" in node_methods
    assert "module1.dummy_layer_mutation" not in node_methods


def test_module_dict_access_methods():
    """Test the various methods to access modules in ModuleDict."""

    class DummyEvolvableModule(EvolvableModule):
        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create a ModuleDict
    module_dict = ModuleDict(device="cpu")

    # Add modules after initialization
    module_dict["module1"] = DummyEvolvableModule(device="cpu")
    module_dict["module2"] = DummyEvolvableModule(device="cpu")

    # Test __getitem__
    assert isinstance(module_dict["module1"], DummyEvolvableModule)

    # Test values()
    values = list(module_dict.values())
    assert len(values) == 2
    assert all(isinstance(module, DummyEvolvableModule) for module in values)

    # Test items()
    items = list(module_dict.items())
    assert len(items) == 2
    assert items[0][0] == "module1" or items[1][0] == "module1"
    assert all(isinstance(item[1], DummyEvolvableModule) for item in items)

    # Test modules()
    modules = module_dict.modules()
    assert len(modules) == 2
    assert "module1" in modules
    assert "module2" in modules


######### Test EvolvableWrapper #########
def test_evolvable_wrapper_initialization():
    """Test EvolvableWrapper initialization with an evolvable module."""

    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_node_mutation(self):
            return {"mutation": "node"}

        @mutation(MutationType.LAYER)
        def dummy_layer_mutation(self):
            return {"mutation": "layer"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create an evolvable module
    module = DummyEvolvableModule(device="cpu")

    # Wrap the module
    wrapper = EvolvableWrapper(module)

    # Check that the wrapper contains the module
    assert wrapper.wrapped == module
    assert wrapper.device == module.device


def test_evolvable_wrapper_mutation_methods():
    """Test that EvolvableWrapper correctly exposes mutation methods from wrapped module."""

    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_node_mutation(self):
            return {"mutation": "node"}

        @mutation(MutationType.LAYER)
        def dummy_layer_mutation(self):
            return {"mutation": "layer"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create an evolvable module
    module = DummyEvolvableModule(device="cpu")

    # Wrap the module
    wrapper = EvolvableWrapper(module)

    # Check that mutation methods are correctly exposed
    wrapper_methods = wrapper.get_mutation_methods()

    # Check that all expected mutation methods are present
    assert "dummy_node_mutation" in wrapper_methods
    assert "dummy_layer_mutation" in wrapper_methods

    # Check that mutation types are preserved
    assert wrapper_methods["dummy_node_mutation"]._mutation_type == MutationType.NODE
    assert wrapper_methods["dummy_layer_mutation"]._mutation_type == MutationType.LAYER

    # Check that the wrapped module's mutations are disabled
    assert not any("_wrapped" in method for method in wrapper.mutation_methods)


class MockMethod:
    _mutation_type = MutationType.NODE
    _recreate_kwargs = {}

    def __call__(self, *args, **kwargs):
        return {"mutation": "mock"}


def test_evolvable_wrapper_mutation_execution():
    """Test that mutations on the wrapper are correctly executed."""
    # In EvolvableWrapper, the wrapper takes over the mutation methods
    # and the original module's methods are disabled

    class DummyEvolvableModule(EvolvableModule):
        @mutation(MutationType.NODE)
        def dummy_mutation(self):
            return {"mutation": "dummy"}

        def forward(self, x):
            pass

        def recreate_network(self):
            pass

    # Create an evolvable module with a mock recreate_network
    module = DummyEvolvableModule(device="cpu")

    # Wrap the module
    wrapper = EvolvableWrapper(module)

    # The wrapper should have adopted the method but the original method is disabled
    # We need to mock it in the wrapper directly
    wrapper.dummy_mutation = MockMethod()

    # Call the mutation method on the wrapper
    result = wrapper.dummy_mutation()

    # Verify the result
    assert result == {"mutation": "mock"}


def test_evolvable_wrapper_forward():
    """Test the forward method being correctly proxied to the wrapped module."""

    class DummyEvolvableModule(EvolvableModule):
        def forward(self, x):
            return "forwarded"

        def recreate_network(self):
            pass

    # Create an evolvable module
    module = DummyEvolvableModule(device="cpu")

    # Wrap the module
    wrapper = EvolvableWrapper(module)

    # Define a forward method that calls the wrapped module
    def forward(self, x):
        return self.wrapped.forward(x)

    # Add the forward method to the wrapper
    wrapper.forward = forward.__get__(wrapper)

    # Call the forward method
    result = wrapper.forward("input")

    # Verify the result
    assert result == "forwarded"



import math
from typing import Optional, Union

import numpy as np
import torch
import torch.nn.functional as F
from gymnasium import spaces

from agilerl.modules.base import EvolvableModule, EvolvableWrapper
from agilerl.typing import ArrayOrTensor, DeviceType, NetConfigType

# NOTE: we still import Normal / Bernoulli solely for continuous & binary helpers,
#       but no Categorical objects are ever instantiated any more.


def apply_action_mask_discrete(
    logits: torch.Tensor, mask: torch.Tensor
) -> torch.Tensor:
    """Apply a mask to the logits.

    :param logits: Logits.
    :type logits: torch.Tensor
    :param mask: Mask.
    :type mask: torch.Tensor
    :return: Logits with mask applied.
    :rtype: torch.Tensor
    """
    return torch.where(mask, logits, torch.full_like(logits, -1e8).to(logits.device))


class TorchDistribution:
    """
    Lightweight distribution‑like helper.
    *   keeps only **raw tensors** (``logits`` or ``mu``/``log_std``)
    *   implements ``sample``, ``log_prob`` and ``entropy`` with pure tensor ops
        → no Python allocations per call, all kernels run on GPU.

    :param action_space: Action space of the environment.
    :type action_space: spaces.Space
    :param logits: Logits.
    :type logits: torch.Tensor
    :param mu: Mean.
    :type mu: torch.Tensor
    :param log_std: Log standard deviation.
    :type log_std: torch.Tensor
    :param squash_output: Whether to squash the output to the action space.
    :type squash_output: bool

    """

    def __init__(
        self,
        *,
        action_space: spaces.Space,
        logits: (
            torch.Tensor | None
        ) = None,  # for discrete / multidiscrete / multibinary
        mu: torch.Tensor | None = None,  # for Box
        log_std: torch.Tensor | None = None,
        squash_output: bool = False,
    ):
        self.action_space = action_space
        self.logits, self.mu, self.log_std = logits, mu, log_std
        self.squash_output = squash_output and isinstance(action_space, spaces.Box)
        self._sampled_action: torch.Tensor | None = None

    # ------------------------------------------------------------------ #
    # fast tensor‑only primitives                                        #
    # ------------------------------------------------------------------ #
    def sample(self) -> torch.Tensor:
        if isinstance(self.action_space, spaces.Discrete):
            probs = torch.softmax(self.logits, dim=-1)
            self._sampled_action = torch.multinomial(probs, 1).squeeze(-1)
            return self._sampled_action

        if isinstance(self.action_space, spaces.Box):
            eps = torch.randn_like(self.mu)
            out = self.mu + torch.exp(self.log_std) * eps
            if self.squash_output:
                out = torch.tanh(out)
            self._sampled_action = out
            return out

        # -------- MultiDiscrete --------
        if isinstance(self.action_space, spaces.MultiDiscrete):
            actions = []
            offset = 0
            for size in self.action_space.nvec:
                logits_i = self.logits[:, offset : offset + size]
                probs_i = torch.softmax(logits_i, dim=-1)
                act_i = torch.multinomial(probs_i, 1).squeeze(-1)
                actions.append(act_i)
                offset += size
            self._sampled_action = torch.stack(actions, dim=-1)
            return self._sampled_action

        # -------- MultiBinary --------
        if isinstance(self.action_space, spaces.MultiBinary):
            probs = torch.sigmoid(self.logits)
            self._sampled_action = torch.bernoulli(
                probs
            )  # Ensures float tensor, removed .to(torch.int64)
            return self._sampled_action

        raise NotImplementedError("Unsupported action space in fast path.")

    def log_prob(self, action: torch.Tensor) -> torch.Tensor:
        if isinstance(self.action_space, spaces.Discrete):
            log_p_all = torch.log_softmax(self.logits, dim=-1)  # Shape (B, N_actions)
            action_long = action.long()

            action_indices_for_gather: torch.Tensor

            if action_long.ndim == log_p_all.ndim - 1:  # action_long is (B,)
                action_indices_for_gather = action_long.unsqueeze(
                    -1
                )  # Converts to (B,1)
            elif action_long.ndim == log_p_all.ndim:  # action_long is (B, K)
                if action_long.shape[-1] == 1:  # action_long is (B,1)
                    action_indices_for_gather = action_long
                elif (
                    action_long.shape == log_p_all.shape
                    and hasattr(self.action_space, "n")
                    and action_long.shape[-1] == self.action_space.n
                ):
                    # Special handling for test case: action is (B, N_actions) for Discrete(N_actions)
                    # Use argmax to get the action index.
                    action_indices_for_gather = torch.argmax(
                        action_long, dim=-1, keepdim=True
                    )  # Converts (B, N_actions) to (B,1)
                else:
                    raise ValueError(
                        f"Action shape {action.shape} is not compatible with Discrete space. "
                        f"Expected (batch_size,), (batch_size, 1), or (batch_size, num_actions) for argmax case. "
                        f"Logits shape: {log_p_all.shape}. Action space: {self.action_space}"
                    )
            else:
                raise ValueError(
                    f"Action tensor ndim {action.ndim} is not compatible with Discrete space logits ndim {log_p_all.ndim}. "
                    f"Expected action ndim to be {log_p_all.ndim-1} or {log_p_all.ndim}."
                )

            return log_p_all.gather(-1, action_indices_for_gather).squeeze(-1)

        if isinstance(self.action_space, spaces.Box):
            var = torch.exp(2 * self.log_std)
            return (
                -0.5
                * (
                    ((action - self.mu) ** 2) / var
                    + 2 * self.log_std
                    + math.log(2 * math.pi)
                )
            ).sum(-1)

        # -------- MultiDiscrete --------
        if isinstance(self.action_space, spaces.MultiDiscrete):
            logps = []
            offset = 0
            for idx, size in enumerate(self.action_space.nvec):
                logits_i = self.logits[:, offset : offset + size]
                logp_all = torch.log_softmax(logits_i, dim=-1)
                act_i = action[:, idx].long()
                logp_i = logp_all.gather(-1, act_i.unsqueeze(-1)).squeeze(-1)
                logps.append(logp_i)
                offset += size
            return torch.stack(logps, dim=-1).sum(-1)

        # -------- MultiBinary --------
        if isinstance(self.action_space, spaces.MultiBinary):
            # log σ(x)  and log (1‑σ(x))
            log_p1 = -F.softplus(-self.logits)
            log_p0 = -self.logits + log_p1
            a = (
                action.float()
            )  # Action for MultiBinary is expected to be float (0.0 or 1.0)
            return (a * log_p1 + (1.0 - a) * log_p0).sum(-1)

        raise NotImplementedError

    def entropy(self) -> torch.Tensor:
        if isinstance(self.action_space, spaces.Discrete):
            p = torch.softmax(self.logits, dim=-1)
            return -(p * torch.log(p + 1e-8)).sum(-1)

        if isinstance(self.action_space, spaces.Box):
            return 0.5 * (1 + math.log(2 * math.pi)) * self.mu.size(
                -1
            ) + self.log_std.sum(-1)

        # -------- MultiDiscrete --------
        if isinstance(self.action_space, spaces.MultiDiscrete):
            entropies = []
            offset = 0
            for size in self.action_space.nvec:
                logits_i = self.logits[:, offset : offset + size]
                p_i = torch.softmax(logits_i, dim=-1)
                ent_i = -(p_i * torch.log(p_i + 1e-8)).sum(-1)
                entropies.append(ent_i)
                offset += size
            return torch.stack(entropies, dim=-1).sum(-1)

        # -------- MultiBinary --------
        if isinstance(self.action_space, spaces.MultiBinary):
            p = torch.sigmoid(self.logits)
            return -(p * torch.log(p + 1e-8) + (1 - p) * torch.log(1 - p + 1e-8)).sum(
                -1
            )

        raise NotImplementedError


class EvolvableDistribution(EvolvableWrapper):
    """Wrapper to output a distribution over an action space for an evolvable module. It provides methods
    to sample actions and compute log probabilities, relevant for many policy-gradient algorithms such as
    PPO, A2C, TRPO.

    :param action_space: Action space of the environment.
    :type action_space: spaces.Space
    :param network: Network that outputs the logits of the distribution.
    :type network: EvolvableModule
    :param action_std_init: Initial log standard deviation of the action distribution. Defaults to 0.0.
    :type action_std_init: float
    :param squash_output: Whether to squash the output to the action space.
    :type squash_output: bool
    :param device: Device to use for the network.
    :type device: DeviceType
    """

    wrapped: EvolvableModule
    dist: Optional[TorchDistribution]
    mask: Optional[ArrayOrTensor]
    log_std: Optional[torch.nn.Parameter]

    def __init__(
        self,
        action_space: spaces.Space,
        network: EvolvableModule,
        action_std_init: float = 0.0,
        squash_output: bool = False,
        device: DeviceType = "cpu",
    ):
        super().__init__(network)

        self.action_space = action_space
        self.action_dim = spaces.flatdim(action_space)
        self.action_std_init = action_std_init
        self.device = device
        self.squash_output = squash_output and isinstance(action_space, spaces.Box)
        self.dist = None
        self.mask = None

        # For continuous action spaces, we also learn the standard
        # deviation (log_std) of the action distribution
        if isinstance(action_space, spaces.Box):
            self.log_std = torch.nn.Parameter(
                torch.ones(1, np.prod(action_space.shape), device=device)
                * action_std_init
            )

    @property
    def net_config(self) -> NetConfigType:
        """Configuration of the network.

        :return: Configuration of the network.
        :rtype: NetConfigType
        """
        return self.wrapped.net_config

    def get_distribution(self, logits: torch.Tensor) -> TorchDistribution:
        """Get the distribution over the action space given an observation.

        :param logits: Output of the network, either logits or probabilities.
        :type logits: torch.Tensor
        :return: Distribution over the action space.
        :rtype: Distribution # This should ideally be TorchDistribution, but keeping for consistency with old file if Distribution was a type alias
        """
        # Normal distribution for Continuous action spaces
        if isinstance(self.action_space, spaces.Box):
            log_std = self.log_std.expand_as(logits)
            # Pass mu and log_std directly to TorchDistribution
            return TorchDistribution(
                action_space=self.action_space,
                mu=logits,
                log_std=log_std,
                squash_output=self.squash_output,
            )

        # Categorical distribution for Discrete action spaces
        elif isinstance(self.action_space, spaces.Discrete):
            # Pass logits directly to TorchDistribution
            return TorchDistribution(
                action_space=self.action_space,
                logits=logits,
                squash_output=self.squash_output,  # squash_output is ignored for discrete
            )

        # List of categorical distributions for MultiDiscrete action spaces
        elif isinstance(self.action_space, spaces.MultiDiscrete):
            # Pass logits directly to TorchDistribution
            return TorchDistribution(
                action_space=self.action_space,
                logits=logits,
                squash_output=self.squash_output,  # squash_output is ignored for discrete
            )

        # Bernoulli distribution for MultiBinary action spaces
        elif isinstance(self.action_space, spaces.MultiBinary):
            # Pass logits directly to TorchDistribution
            return TorchDistribution(
                action_space=self.action_space,
                logits=logits,
                squash_output=self.squash_output,  # squash_output is ignored for discrete
            )
        else:
            raise NotImplementedError(
                f"Action space {self.action_space} not supported."
            )

    def log_prob(self, action: torch.Tensor) -> torch.Tensor:
        """Get the log probability of the action.

        :param action: Action.
        :type action: torch.Tensor
        :return: Log probability of the action.
        :rtype: torch.Tensor
        """
        if self.dist is None:
            raise ValueError("Distribution not initialized. Call forward first.")

        # The new TorchDistribution handles squashing correction internally for Box space
        return self.dist.log_prob(action)

    def entropy(self) -> torch.Tensor:
        """Get the entropy of the action distribution.

        :return: Entropy of the action distribution.
        :rtype: torch.Tensor
        """
        if self.dist is None:
            raise ValueError("Distribution not initialized. Call forward first.")

        # The new TorchDistribution returns analytical entropy for supported spaces
        return self.dist.entropy()

    def apply_mask(self, logits: torch.Tensor, mask: ArrayOrTensor) -> torch.Tensor:
        """Apply a mask to the logits.

        :param logits: Logits.
        :type logits: torch.Tensor
        :param mask: Mask.
        :type mask: ArrayOrTensor
        :return: Logits with mask applied.
        :rtype: torch.Tensor
        """
        # Convert mask to tensor and reshape to match logits shape
        mask = torch.as_tensor(mask, dtype=torch.bool, device=self.device).view(
            logits.shape
        )

        if isinstance(self.action_space, spaces.Discrete):
            masked_logits = apply_action_mask_discrete(logits, mask)
        elif isinstance(self.action_space, (spaces.MultiDiscrete, spaces.MultiBinary)):
            splits = (
                list(self.action_space.nvec)
                if isinstance(self.action_space, spaces.MultiDiscrete)
                else [
                    self.action_space.n
                ]  # For MultiBinary, nvec is not present, use n
            )
            # Split mask and logits into separate distributions
            split_masks = torch.split(mask, splits, dim=1)
            split_logits = torch.split(logits, splits, dim=1)

            # Apply mask to each split
            masked_logits = []
            for split_logits_i, split_mask_i in zip(
                split_logits, split_masks
            ):  # Renamed for clarity
                masked_logits.append(
                    apply_action_mask_discrete(split_logits_i, split_mask_i)
                )

            masked_logits = torch.cat(masked_logits, dim=1)
        else:
            # This should ideally not be reached if get_distribution handles the space,
            # but keeping for safety.
            raise NotImplementedError(
                f"Action space {self.action_space} not supported for masking."
            )

        return masked_logits

    def forward(
        self,
        latent: torch.Tensor,
        action_mask: Optional[ArrayOrTensor] = None,
        sample: bool = True,
    ) -> Union[
        tuple[torch.Tensor, torch.Tensor, torch.Tensor], tuple[None, None, torch.Tensor]
    ]:
        """Forward pass of the network.

        :param latent: Latent space representation.
        :type latent: torch.Tensor
        :param action_mask: Mask to apply to the logits. Defaults to None.
        :type action_mask: Optional[ArrayOrTensor]
        :param sample: Whether to sample an action or return the mode/mean. Defaults to True.
        :type sample: bool
        :return: Action and log probability of the action.
        :rtype: Union[tuple[torch.Tensor, torch.Tensor, torch.Tensor], tuple[None, torch.Tensor, torch.Tensor]]
        """
        logits = self.wrapped(latent)

        if action_mask is not None:
            if isinstance(action_mask, (np.ndarray, list)):
                # Attempt to stack if it's a list of arrays or object array, typical for vectorized envs
                if isinstance(action_mask, list) or (
                    isinstance(action_mask, np.ndarray)
                    and action_mask.dtype == np.object_
                ):
                    try:
                        action_mask = np.stack(action_mask)
                    except Exception:
                        # If stacking fails, it might be a non-uniform list or other structure not directly convertible.
                        # This path assumes action_mask should become a single tensor.
                        # If it's already a correct tensor, as_tensor below handles it.
                        pass  # Allow as_tensor to handle or raise error if still problematic

            # Ensure action_mask is a tensor before applying.
            # The view in apply_mask expects a compatible shape or will error.
            action_mask = torch.as_tensor(
                action_mask, device=self.device, dtype=torch.bool
            )

            logits = self.apply_mask(logits, action_mask)

        # Distribution from logits
        # get_distribution now creates the new TorchDistribution object
        self.dist = self.get_distribution(logits)

        # Sample action, compute log probability and entropy
        if sample:
            action = self.dist.sample()
            log_prob = self.dist.log_prob(action)
        else:
            action = None  # Mode/mean might be more appropriate if not sampling
            log_prob = (
                None  # Log prob of mode/mean typically not used in PPO sample step
            )

        entropy = self.dist.entropy()
        return action, log_prob, entropy

    def clone(self) -> "EvolvableDistribution":
        """Clones the distribution.

        :return: Cloned distribution.
        :rtype: EvolvableDistribution
        """
        clone = EvolvableDistribution(
            action_space=self.action_space,
            network=self.wrapped.clone(),
            action_std_init=self.action_std_init,
            squash_output=self.squash_output,
            device=self.device,
        )
        clone.rng = self.rng
        return clone



import copy

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces
from gymnasium.spaces import Discrete
from torch._dynamo import OptimizedModule

from agilerl.algorithms.matd3 import MATD3
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput, ModuleDict
from agilerl.modules.custom_components import GumbelSoftmax
from agilerl.networks.actors import DeterministicActor
from agilerl.networks.q_networks import ContinuousQNetwork
from agilerl.utils.algo_utils import concatenate_spaces
from agilerl.utils.evolvable_networks import get_default_encoder_config
from agilerl.utils.utils import make_multi_agent_vect_envs
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import assert_not_equal_state_dict, assert_state_dicts_equal
from tests.test_algorithms.test_multi_agent.test_maddpg import DummyMultiEnv


class MultiAgentCNNActor(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv3d(
            in_channels=3, out_channels=16, kernel_size=(1, 3, 3), stride=4
        )
        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=(1, 3, 3), stride=2
        )
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(288, 256)
        self.fc2 = nn.Linear(256, 2)
        self.relu = nn.ReLU()
        self.output_activation = GumbelSoftmax()

    def forward(self, state_tensor):
        x = self.relu(self.conv1(state_tensor))
        x = self.relu(self.conv2(x))
        x = x.view(x.size(0), -1)
        x = self.relu(self.fc1(x))
        x = self.output_activation(self.fc2(x))

        return x


class MultiAgentCNNCritic(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv3d(
            in_channels=3, out_channels=16, kernel_size=(2, 3, 3), stride=4
        )
        self.conv2 = nn.Conv3d(
            in_channels=16, out_channels=32, kernel_size=(1, 3, 3), stride=2
        )
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(290, 256)
        self.fc2 = nn.Linear(256, 2)
        self.relu = nn.ReLU()

    def forward(self, state_tensor, action_tensor):
        x = self.relu(self.conv1(state_tensor))
        x = self.relu(self.conv2(x))
        x = x.view(x.size(0), -1)
        x = torch.cat([x, action_tensor], dim=1)
        x = self.relu(self.fc1(x))
        x = self.fc2(x)

        return x


class DummyContinuousQNetwork(ContinuousQNetwork):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    def no_sync(self):
        class DummyNoSync:
            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_value, traceback):
                pass  # Add cleanup or handling if needed

        return DummyNoSync()


class DummyDeterministicActor(DeterministicActor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return super().forward(*args, **kwargs)

    def no_sync(self):
        class DummyNoSync:
            def __enter__(self):
                return self

            def __exit__(self, exc_type, exc_value, traceback):
                pass  # Add cleanup or handling if needed

        return DummyNoSync()


@pytest.fixture(scope="function")
def mlp_actor(observation_spaces, action_spaces, request):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    return nn.Sequential(
        nn.Linear(observation_spaces[0].shape[0], 64),
        nn.ReLU(),
        nn.Linear(64, action_spaces[0].n),
        GumbelSoftmax(),
    )


@pytest.fixture(scope="function")
def mlp_critic(action_spaces, observation_spaces, request):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    return nn.Sequential(
        nn.Linear(observation_spaces[0].shape[0] + action_spaces[0].n, 64),
        nn.ReLU(),
        nn.Linear(64, 1),
    )


@pytest.fixture(scope="function")
def cnn_actor():
    return MultiAgentCNNActor()


@pytest.fixture(scope="function")
def cnn_critic():
    return MultiAgentCNNCritic()


@pytest.fixture(scope="function")
def accelerated_experiences(
    batch_size, observation_spaces, action_spaces, agent_ids, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    one_hot = all(isinstance(space, Discrete) for space in observation_spaces)
    discrete_actions = all(isinstance(space, Discrete) for space in action_spaces)
    state_size = (
        observation_spaces[0].shape if not one_hot else (observation_spaces[0].n,)
    )
    action_size = action_spaces[0].n if discrete_actions else action_spaces[0].shape[0]
    if one_hot:
        states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float()
            for agent in agent_ids
        }
    else:
        states = {agent: torch.randn(batch_size, *state_size) for agent in agent_ids}

    actions = {agent: torch.randn(batch_size, action_size) for agent in agent_ids}
    rewards = {agent: torch.randn(batch_size, 1) for agent in agent_ids}
    dones = {agent: torch.randint(0, 2, (batch_size, 1)) for agent in agent_ids}
    if one_hot:
        next_states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float()
            for agent in agent_ids
        }
    else:
        next_states = {
            agent: torch.randn(batch_size, *state_size) for agent in agent_ids
        }

    yield states, actions, rewards, next_states, dones


@pytest.fixture(scope="function")
def experiences(
    batch_size, observation_spaces, action_spaces, agent_ids, device, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    one_hot = all(isinstance(space, Discrete) for space in observation_spaces)
    discrete_actions = all(isinstance(space, Discrete) for space in action_spaces)
    state_size = (
        observation_spaces[0].shape if not one_hot else (observation_spaces[0].n,)
    )
    action_size = action_spaces[0].n if discrete_actions else action_spaces[0].shape[0]
    if one_hot:
        states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float().to(device)
            for agent in agent_ids
        }
    else:
        states = {
            agent: torch.randn(batch_size, *state_size).to(device)
            for agent in agent_ids
        }

    actions = {
        agent: torch.randn(batch_size, action_size).to(device) for agent in agent_ids
    }
    rewards = {agent: torch.randn(batch_size, 1).to(device) for agent in agent_ids}
    dones = {
        agent: torch.randint(0, 2, (batch_size, 1)).to(device) for agent in agent_ids
    }
    if one_hot:
        next_states = {
            agent: torch.randint(0, state_size[0], (batch_size, 1)).float().to(device)
            for agent in agent_ids
        }
    else:
        next_states = {
            agent: torch.randn(batch_size, *state_size).to(device)
            for agent in agent_ids
        }

    yield states, actions, rewards, next_states, dones


@pytest.mark.parametrize(
    "observation_spaces",
    [
        "ma_vector_space",
        "ma_image_space",
        "ma_discrete_space",
        "ma_multidiscrete_space",
        "ma_dict_space",
    ],
)
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_matd3_with_net_config(
    observation_spaces, ma_vector_space, accelerator_flag, device, compile_mode, request
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    net_config = {
        "encoder_config": get_default_encoder_config(observation_spaces[0]),
        "head_config": {"hidden_size": [16]},
    }
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    expl_noise = 0.1
    batch_size = 64
    policy_freq = 2
    accelerator = Accelerator() if accelerator_flag else None
    matd3 = MATD3(
        observation_spaces=observation_spaces,
        action_spaces=ma_vector_space,
        net_config=net_config,
        agent_ids=agent_ids,
        accelerator=accelerator,
        device=device,
        policy_freq=policy_freq,
        torch_compiler=compile_mode,
    )
    assert matd3.observation_spaces == observation_spaces
    assert matd3.action_spaces == ma_vector_space
    assert matd3.policy_freq == policy_freq
    assert matd3.n_agents == len(agent_ids)
    assert matd3.agent_ids == agent_ids
    for noise_vec in matd3.expl_noise.values():
        assert torch.all(noise_vec == expl_noise)

    assert matd3.batch_size == batch_size
    assert matd3.scores == []
    assert matd3.fitness == []
    assert matd3.steps == [0]

    if compile_mode is not None and accelerator is None:
        assert all(
            isinstance(actor, OptimizedModule) for actor in matd3.actors.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in matd3.critics_1.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in matd3.critics_2.values()
        )
    else:
        assert all(
            isinstance(actor, DeterministicActor) for actor in matd3.actors.values()
        )
        assert all(
            isinstance(critic, ContinuousQNetwork)
            for critic in matd3.critics_1.values()
        )
        assert all(
            isinstance(critic, ContinuousQNetwork)
            for critic in matd3.critics_2.values()
        )

    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    for agent_id in matd3.agent_ids:
        actor_optimizer = matd3.actor_optimizers[agent_id]
        critic_1_optimizer = matd3.critic_1_optimizers[agent_id]
        critic_2_optimizer = matd3.critic_2_optimizers[agent_id]
        assert isinstance(actor_optimizer, expected_optimizer_cls)
        assert isinstance(critic_1_optimizer, expected_optimizer_cls)
        assert isinstance(critic_2_optimizer, expected_optimizer_cls)

    assert isinstance(matd3.criterion, nn.MSELoss)


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_initialize_matd3_with_mlp_networks_gumbel_softmax(
    mlp_actor,
    mlp_critic,
    observation_spaces,
    action_spaces,
    device,
    request,
):
    compile_mode = "reduce-overhead"
    net_config = {
        "head_config": {
            "hidden_size": [64, 64],
            "min_hidden_layers": 1,
            "max_hidden_layers": 3,
            "min_mlp_nodes": 64,
            "max_mlp_nodes": 500,
            "output_activation": "GumbelSoftmax",
            "activation": "ReLU",
        }
    }
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    matd3 = MATD3(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        net_config=net_config,
        device=device,
        torch_compiler=compile_mode,
    )
    assert matd3.torch_compiler == "default"


# TODO: This will be deprecated in the future
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_initialize_matd3_with_mlp_networks(
    mlp_actor,
    mlp_critic,
    observation_spaces,
    action_spaces,
    accelerator_flag,
    device,
    compile_mode,
    request,
):
    accelerator = Accelerator() if accelerator_flag else None
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    evo_actors = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=mlp_actor,
                input_tensor=torch.randn(1, observation_spaces[0].shape[0]),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics_1 = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=mlp_critic,
                input_tensor=torch.randn(
                    1, observation_spaces[0].shape[0] + action_spaces[0].n
                ),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics_2 = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=mlp_critic,
                input_tensor=torch.randn(
                    1, observation_spaces[0].shape[0] + action_spaces[0].n
                ),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics = [evo_critics_1, evo_critics_2]
    matd3 = MATD3(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        accelerator=accelerator,
        policy_freq=2,
        torch_compiler=compile_mode,
    )
    expected_module_cls = (
        OptimizedModule
        if compile_mode is not None and accelerator is None
        else MakeEvolvable
    )
    assert all(
        isinstance(actor, expected_module_cls) for actor in matd3.actors.values()
    )
    assert all(
        isinstance(critic, expected_module_cls) for critic in matd3.critics_1.values()
    )
    assert all(
        isinstance(critic, expected_module_cls) for critic in matd3.critics_2.values()
    )

    assert matd3.observation_spaces == observation_spaces
    assert matd3.action_spaces == action_spaces
    assert matd3.n_agents == len(agent_ids)
    assert matd3.policy_freq == 2
    assert matd3.agent_ids == agent_ids
    assert matd3.scores == []
    assert matd3.fitness == []
    assert matd3.steps == [0]

    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    for agent_id in matd3.agent_ids:
        actor_optimizer = matd3.actor_optimizers[agent_id]
        critic_1_optimizer = matd3.critic_1_optimizers[agent_id]
        critic_2_optimizer = matd3.critic_2_optimizers[agent_id]
        assert isinstance(actor_optimizer, expected_optimizer_cls)
        assert isinstance(critic_1_optimizer, expected_optimizer_cls)
        assert isinstance(critic_2_optimizer, expected_optimizer_cls)

    assert isinstance(matd3.criterion, nn.MSELoss)


# TODO: This will be deprecated in the future
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_matd3_with_cnn_networks(
    cnn_actor,
    cnn_critic,
    ma_image_space,
    ma_discrete_space,
    accelerator_flag,
    device,
    compile_mode,
):
    accelerator = Accelerator() if accelerator_flag else None
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    evo_actors = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=cnn_actor,
                input_tensor=torch.randn(1, 3, 1, 32, 32),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics_1 = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=cnn_critic,
                input_tensor=torch.randn(1, 3, 3, 32, 32),
                secondary_input_tensor=torch.randn(1, 2),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics_2 = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=cnn_critic,
                input_tensor=torch.randn(1, 3, 3, 32, 32),
                secondary_input_tensor=torch.randn(1, 2),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics = [evo_critics_1, evo_critics_2]
    matd3 = MATD3(
        observation_spaces=ma_image_space,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        accelerator=accelerator,
        policy_freq=2,
        torch_compiler=compile_mode,
    )
    expected_module_cls = (
        OptimizedModule
        if compile_mode is not None and accelerator is None
        else MakeEvolvable
    )
    assert all(
        isinstance(actor, expected_module_cls) for actor in matd3.actors.values()
    )
    assert all(
        isinstance(critic, expected_module_cls) for critic in matd3.critics_1.values()
    )
    assert all(
        isinstance(critic, expected_module_cls) for critic in matd3.critics_2.values()
    )
    assert matd3.observation_spaces == ma_image_space
    assert matd3.policy_freq == 2
    assert matd3.action_spaces == ma_discrete_space
    assert matd3.n_agents == len(agent_ids)
    assert matd3.agent_ids == agent_ids
    assert matd3.scores == []
    assert matd3.fitness == []
    assert matd3.steps == [0]

    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    for agent_id in matd3.agent_ids:
        actor_optimizer = matd3.actor_optimizers[agent_id]
        critic_1_optimizer = matd3.critic_1_optimizers[agent_id]
        critic_2_optimizer = matd3.critic_2_optimizers[agent_id]
        assert isinstance(actor_optimizer, expected_optimizer_cls)
        assert isinstance(critic_1_optimizer, expected_optimizer_cls)
        assert isinstance(critic_2_optimizer, expected_optimizer_cls)

    assert isinstance(matd3.criterion, nn.MSELoss)


@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize(
    "observation_spaces, encoder_cls",
    [
        ("ma_vector_space", EvolvableMLP),
        ("ma_image_space", EvolvableCNN),
    ],
)
def test_initialize_matd3_with_evo_networks(
    observation_spaces,
    ma_discrete_space,
    encoder_cls,
    device,
    compile_mode,
    accelerator,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_space = spaces.Dict(
        {agent_id: observation_spaces[idx] for idx, agent_id in enumerate(agent_ids)}
    )

    evo_actors = ModuleDict(
        {
            agent_id: DeterministicActor(
                observation_spaces[idx], ma_discrete_space[idx], device=device
            )
            for idx, agent_id in enumerate(agent_ids)
        }
    )
    evo_critics_1 = ModuleDict(
        {
            agent_id: ContinuousQNetwork(
                observation_space=observation_space,
                action_space=concatenate_spaces(ma_discrete_space),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics_2 = ModuleDict(
        {
            agent_id: ContinuousQNetwork(
                observation_space=observation_space,
                action_space=concatenate_spaces(ma_discrete_space),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    evo_critics = [evo_critics_1, evo_critics_2]
    matd3 = MATD3(
        observation_spaces=observation_spaces,
        action_spaces=ma_discrete_space,
        agent_ids=agent_ids,
        actor_networks=evo_actors,
        critic_networks=evo_critics,
        device=device,
        torch_compiler=compile_mode,
        accelerator=accelerator,
    )
    if compile_mode is not None and accelerator is None:
        assert all(
            isinstance(actor, OptimizedModule) for actor in matd3.actors.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in matd3.critics_1.values()
        )
        assert all(
            isinstance(critic, OptimizedModule) for critic in matd3.critics_2.values()
        )
    else:
        assert all(
            isinstance(actor.encoder, encoder_cls) for actor in matd3.actors.values()
        )
        assert all(
            isinstance(critic.encoder, EvolvableMultiInput)
            for critic in matd3.critics_1.values()
        )
        assert all(
            isinstance(critic.encoder, EvolvableMultiInput)
            for critic in matd3.critics_2.values()
        )
    assert matd3.observation_spaces == observation_spaces
    assert matd3.policy_freq == 2
    assert matd3.action_spaces == ma_discrete_space
    assert matd3.n_agents == len(agent_ids)
    assert matd3.agent_ids == agent_ids
    assert matd3.scores == []
    assert matd3.fitness == []
    assert matd3.steps == [0]

    expected_optimizer_cls = optim.Adam if accelerator is None else AcceleratedOptimizer
    for agent_id in matd3.agent_ids:
        actor_optimizer = matd3.actor_optimizers[agent_id]
        critic_1_optimizer = matd3.critic_1_optimizers[agent_id]
        critic_2_optimizer = matd3.critic_2_optimizers[agent_id]
        assert isinstance(actor_optimizer, expected_optimizer_cls)
        assert isinstance(critic_1_optimizer, expected_optimizer_cls)
        assert isinstance(critic_2_optimizer, expected_optimizer_cls)

    assert isinstance(matd3.criterion, nn.MSELoss)


@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_initialize_matd3_with_incorrect_evo_networks(
    compile_mode, ma_vector_space, ma_discrete_space
):
    evo_actors = []
    evo_critics = []
    with pytest.raises(AssertionError):
        _ = MATD3(
            observation_spaces=ma_vector_space,
            action_spaces=ma_discrete_space,
            agent_ids=["agent_0", "agent_1", "other_agent_0"],
            actor_networks=evo_actors,
            critic_networks=evo_critics,
            torch_compiler=compile_mode,
        )


@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space"])
def test_matd3_init_warning(
    mlp_actor, device, compile_mode, observation_spaces, action_spaces, request
):
    warning_string = "Actor and critic network must both be supplied to use custom networks. Defaulting to net config."
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    evo_actors = ModuleDict(
        {
            agent_id: MakeEvolvable(
                network=mlp_actor,
                input_tensor=torch.randn(1, observation_spaces[0].shape[0]),
                device=device,
            )
            for agent_id in agent_ids
        }
    )
    with pytest.warns(UserWarning, match=warning_string):
        MATD3(
            observation_spaces=observation_spaces,
            action_spaces=action_spaces,
            agent_ids=agent_ids,
            actor_networks=evo_actors,
            device=device,
            torch_compiler=compile_mode,
        )


@pytest.mark.parametrize(
    "mode", [None, 0, False, "default", "reduce-overhead", "max-autotune"]
)
def test_matd3_init_with_compile_no_error(mode, ma_vector_space, device):
    matd3 = MATD3(
        observation_spaces=ma_vector_space,
        action_spaces=copy.deepcopy(ma_vector_space),
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        device=device,
        torch_compiler=mode,
    )
    if isinstance(mode, str):
        assert all(isinstance(m, OptimizedModule) for m in matd3.actors.values())
        assert all(isinstance(m, OptimizedModule) for m in matd3.actor_targets.values())
        assert all(isinstance(m, OptimizedModule) for m in matd3.critics_1.values())
        assert all(isinstance(m, OptimizedModule) for m in matd3.critics_2.values())
        assert all(
            isinstance(m, OptimizedModule) for m in matd3.critic_targets_1.values()
        )
        assert all(
            isinstance(m, OptimizedModule) for m in matd3.critic_targets_2.values()
        )
        assert matd3.torch_compiler == mode
    else:
        assert isinstance(matd3, MATD3)


@pytest.mark.parametrize("mode", [1, True, "max-autotune-no-cudagraphs"])
def test_matd3_init_with_compile_error(mode, ma_vector_space, device):
    err_string = (
        "Choose between torch compiler modes: "
        "default, reduce-overhead, max-autotune or None"
    )
    with pytest.raises(AssertionError, match=err_string):
        MATD3(
            observation_spaces=ma_vector_space,
            action_spaces=copy.deepcopy(ma_vector_space),
            agent_ids=["agent_0", "agent_1", "other_agent_0"],
            device=device,
            torch_compiler=mode,
        )


@pytest.mark.parametrize(
    "observation_spaces", ["ma_vector_space", "ma_discrete_space", "ma_image_space"]
)
@pytest.mark.parametrize("action_spaces", ["ma_vector_space", "ma_discrete_space"])
@pytest.mark.parametrize("training", [0, 1])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_matd3_get_action(
    training, observation_spaces, action_spaces, device, compile_mode, request
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    if all(isinstance(space, spaces.Discrete) for space in observation_spaces):
        state = {
            agent: np.random.randint(0, observation_spaces[idx].n, 1)
            for idx, agent in enumerate(agent_ids)
        }
    else:
        state = {
            agent: np.random.randn(*observation_spaces[idx].shape)
            for idx, agent in enumerate(agent_ids)
        }

    matd3 = MATD3(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )
    matd3.set_training_mode(bool(training))
    processed_action, raw_action = matd3.get_action(state)
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    for idx, env_actions in enumerate(list(raw_action.values())):
        action_dim = (
            action_spaces[idx].shape[0]
            if isinstance(action_spaces[idx], spaces.Box)
            else action_spaces[idx].n
        )
        for action in env_actions:
            assert len(action) == action_dim
            if discrete_actions:
                torch.testing.assert_close(
                    sum(action),
                    1.0,
                    atol=0.1,
                    rtol=1e-3,
                )
            assert action.dtype == np.float32
            assert -1 <= action.all() <= 1

    if discrete_actions:
        for idx, env_action in enumerate(list(processed_action.values())):
            for action in env_action:
                assert action <= action_spaces[idx].n - 1
    matd3 = None


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space", "ma_image_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space", "ma_vector_space"])
@pytest.mark.parametrize("training", [0, 1])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_matd3_get_action_distributed(
    training, observation_spaces, action_spaces, compile_mode, request
):
    accelerator = Accelerator()
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    state = {
        agent: np.random.randn(*observation_spaces[idx].shape)
        for idx, agent in enumerate(agent_ids)
    }
    matd3 = MATD3(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        accelerator=accelerator,
        torch_compiler=compile_mode,
    )
    new_actors = ModuleDict(
        {
            agent_id: DummyDeterministicActor(
                observation_space=actor.observation_space,
                action_space=actor.action_space,
                encoder_config=actor.encoder.net_config,
                head_config=actor.head_net.net_config,
                device=actor.device,
            )
            for agent_id, actor in matd3.actors.items()
        }
    )
    matd3.actors = new_actors
    matd3.set_training_mode(bool(training))
    processed_action, raw_action = matd3.get_action(state)
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    for idx, env_actions in enumerate(list(raw_action.values())):
        action_dim = (
            action_spaces[idx].shape[0]
            if isinstance(action_spaces[idx], spaces.Box)
            else action_spaces[idx].n
        )
        for action in env_actions:
            assert len(action) == action_dim
            if discrete_actions:
                torch.testing.assert_close(
                    sum(action),
                    1.0,
                    atol=0.1,
                    rtol=1e-3,
                )
            assert action.dtype == np.float32
            assert -1 <= action.all() <= 1

    if discrete_actions:
        for idx, env_action in enumerate(list(processed_action.values())):
            action_dim = (
                action_spaces[idx].shape[0]
                if isinstance(action_spaces[idx], spaces.Box)
                else action_spaces[idx].n
            )
            for action in env_action:
                assert action <= action_dim - 1


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space", "ma_vector_space"])
@pytest.mark.parametrize("training", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_matd3_get_action_agent_masking(
    training, observation_spaces, action_spaces, device, compile_mode, request
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    state = {
        agent: np.random.randn(*observation_spaces[0].shape) for agent in agent_ids
    }
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    if discrete_actions:
        info = {
            "agent_0": {"env_defined_actions": 1},
            "agent_1": {"env_defined_actions": 1},
            "other_agent_0": {"env_defined_actions": None},
        }
    else:
        info = {
            "agent_0": {"env_defined_actions": np.array([0, 1, 0, 1, 0, 1])},
            "agent_1": {"env_defined_actions": np.array([0, 1, 0, 1, 0, 1])},
            "other_agent_0": {"env_defined_actions": None},
        }
    matd3 = MATD3(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )
    matd3.set_training_mode(training)
    action, _ = matd3.get_action(state, infos=info)
    if discrete_actions:
        assert np.array_equal(action["agent_0"], np.array([1])), action["agent_0"]
    else:
        assert np.array_equal(
            action["agent_0"], np.array([[0, 1, 0, 1, 0, 1]])
        ), action["agent_0"]


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space", "ma_vector_space"])
@pytest.mark.parametrize("training", [False, True])
@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_matd3_get_action_vectorized_agent_masking(
    training, observation_spaces, action_spaces, device, compile_mode, request
):
    num_envs = 6
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    state = {
        agent: np.array(
            [np.random.randn(*observation_spaces[0].shape) for _ in range(num_envs)]
        )
        for agent in agent_ids
    }
    discrete_actions = all(
        isinstance(space, spaces.Discrete) for space in action_spaces
    )
    if discrete_actions:
        env_defined_action = np.array(
            [
                np.random.randint(0, observation_spaces[0].shape[0] + 1)
                for _ in range(num_envs)
            ]
        )
    else:
        env_defined_action = np.array(
            [np.random.randn(*observation_spaces[0].shape) for _ in range(num_envs)]
        )
    nan_array = np.zeros(env_defined_action.shape)
    nan_array[:] = np.nan
    info = {
        "agent_0": {"env_defined_actions": env_defined_action},
        "agent_1": {"env_defined_actions": env_defined_action},
        "other_agent_0": {"env_defined_actions": nan_array},
    }
    matd3 = MATD3(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
        torch_compiler=compile_mode,
    )
    matd3.set_training_mode(training)
    action, _ = matd3.get_action(state, infos=info)
    if discrete_actions:
        assert np.array_equal(
            action["agent_0"].squeeze(), info["agent_0"]["env_defined_actions"]
        ), action["agent_0"]
    else:
        assert np.isclose(
            action["agent_0"], info["agent_0"]["env_defined_actions"]
        ).all(), action["agent_0"]


@pytest.mark.parametrize("training", [False, True])
def test_matd3_get_action_action_masking_exception(
    training, ma_vector_space, ma_discrete_space, device
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: {
            "observation": np.random.randn(*ma_vector_space[idx].shape),
            "action_mask": [0, 1],
        }
        for idx, agent in enumerate(agent_ids)
    }
    matd3 = MATD3(
        ma_vector_space,
        ma_discrete_space,
        agent_ids=agent_ids,
        device=device,
    )
    with pytest.raises(AssertionError):
        matd3.set_training_mode(training)
        _, raw_action = matd3.get_action(state)


@pytest.mark.parametrize("training", [False, True])
def test_matd3_get_action_action_masking(
    training, ma_vector_space, ma_discrete_space, device
):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    state = {
        agent: np.random.randn(*ma_vector_space[idx].shape)
        for idx, agent in enumerate(agent_ids)
    }
    info = {
        agent: {
            "action_mask": [0, 1],
        }
        for idx, agent in enumerate(agent_ids)
    }
    matd3 = MATD3(
        ma_vector_space,
        ma_discrete_space,
        agent_ids=agent_ids,
        device=device,
    )
    matd3.set_training_mode(training)
    action, _ = matd3.get_action(state, info)
    assert all(i in [1, 3] for i in action.values())


@pytest.mark.parametrize(
    "observation_spaces", ["ma_discrete_space", "ma_vector_space", "ma_image_space"]
)
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space", "ma_vector_space"])
@pytest.mark.parametrize("batch_size", [64])
@pytest.mark.parametrize("agent_ids", [["agent_0", "agent_1", "other_agent_0"]])
@pytest.mark.parametrize("compile_mode", [None])
def test_matd3_learns_from_experiences(
    observation_spaces,
    experiences,
    batch_size,
    action_spaces,
    agent_ids,
    device,
    compile_mode,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    policy_freq = 2
    matd3 = MATD3(
        observation_spaces,
        action_spaces,
        agent_ids=agent_ids,
        device=device,
        policy_freq=policy_freq,
        torch_compiler=compile_mode,
    )
    actors_pre_learn_sd = {
        agent_id: copy.deepcopy(actor.state_dict())
        for agent_id, actor in matd3.actors.items()
    }
    critics_1_pre_learn_sd = {
        agent_id: copy.deepcopy(critic_1.state_dict())
        for agent_id, critic_1 in matd3.critics_1.items()
    }
    critics_2_pre_learn_sd = {
        agent_id: copy.deepcopy(critic_2.state_dict())
        for agent_id, critic_2 in matd3.critics_2.items()
    }

    for _ in range(2 * policy_freq):
        matd3.scores.append(0)
        loss = matd3.learn(experiences)

    assert isinstance(loss, dict)

    for agent_id in matd3.agent_ids:
        assert loss[agent_id][-1] >= 0.0

    for agent_id, old_actor_target in matd3.actor_targets.items():
        updated_actor_target = matd3.actor_targets[agent_id]
        assert old_actor_target == updated_actor_target

    for agent_id, old_actor_state_dict in actors_pre_learn_sd.items():
        updated_actor = matd3.actors[agent_id]
        assert_not_equal_state_dict(old_actor_state_dict, updated_actor.state_dict())

    for agent_id, old_critic_target in matd3.critic_targets_1.items():
        updated_critic_target = matd3.critic_targets_1[agent_id]
        assert old_critic_target == updated_critic_target

    for agent_id, old_critic_state_dict in critics_1_pre_learn_sd.items():
        updated_critic = matd3.critics_1[agent_id]
        assert_not_equal_state_dict(old_critic_state_dict, updated_critic.state_dict())

    for agent_id, old_critic_target in matd3.critic_targets_2.items():
        updated_critic_target = matd3.critic_targets_2[agent_id]
        assert old_critic_target == updated_critic_target

    for agent_id, old_critic_state_dict in critics_2_pre_learn_sd.items():
        updated_critic = matd3.critics_2[agent_id]
        assert_not_equal_state_dict(old_critic_state_dict, updated_critic.state_dict())


def no_sync(self):
    class DummyNoSync:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            pass  # Add cleanup or handling if needed

    return DummyNoSync()


@pytest.mark.parametrize(
    "observation_spaces", ["ma_vector_space", "ma_discrete_space", "ma_image_space"]
)
@pytest.mark.parametrize("action_spaces", ["ma_discrete_space", "ma_vector_space"])
@pytest.mark.parametrize("batch_size", [64])
@pytest.mark.parametrize("agent_ids", [["agent_0", "agent_1", "other_agent_0"]])
@pytest.mark.parametrize("compile_mode", [None])
def test_matd3_learns_from_experiences_distributed(
    observation_spaces,
    accelerated_experiences,
    batch_size,
    action_spaces,
    agent_ids,
    compile_mode,
    request,
):
    accelerator = Accelerator(device_placement=False)
    observation_spaces = request.getfixturevalue(observation_spaces)
    action_spaces = request.getfixturevalue(action_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    policy_freq = 2
    matd3 = MATD3(
        observation_spaces=observation_spaces,
        action_spaces=action_spaces,
        agent_ids=agent_ids,
        accelerator=accelerator,
        policy_freq=policy_freq,
        torch_compiler=compile_mode,
    )

    for agent_id in matd3.agent_ids:
        actor = matd3.actors[agent_id]
        critic_1 = matd3.critics_1[agent_id]
        critic_2 = matd3.critics_2[agent_id]
        actor_target = matd3.actor_targets[agent_id]
        critic_target_1 = matd3.critic_targets_1[agent_id]
        critic_target_2 = matd3.critic_targets_2[agent_id]
        actor.no_sync = no_sync.__get__(actor)
        critic_1.no_sync = no_sync.__get__(critic_1)
        critic_2.no_sync = no_sync.__get__(critic_2)
        actor_target.no_sync = no_sync.__get__(actor_target)
        critic_target_1.no_sync = no_sync.__get__(critic_target_1)
        critic_target_2.no_sync = no_sync.__get__(critic_target_2)

    actors_pre_learn_sd = {
        agent_id: copy.deepcopy(actor.state_dict())
        for agent_id, actor in matd3.actors.items()
    }
    critics_1_pre_learn_sd = {
        agent_id: copy.deepcopy(critic_1.state_dict())
        for agent_id, critic_1 in matd3.critics_1.items()
    }
    critics_2_pre_learn_sd = {
        agent_id: copy.deepcopy(critic_2.state_dict())
        for agent_id, critic_2 in matd3.critics_2.items()
    }

    for _ in range(2 * policy_freq):
        matd3.scores.append(0)
        loss = matd3.learn(accelerated_experiences)

    assert isinstance(loss, dict)
    for agent_id in matd3.agent_ids:
        assert loss[agent_id][-1] >= 0.0

    for agent_id, old_actor_sd in actors_pre_learn_sd.items():
        updated_actor = matd3.actors[agent_id]
        assert_not_equal_state_dict(old_actor_sd, updated_actor.state_dict())

    for agent_id, old_critic_1_sd in critics_1_pre_learn_sd.items():
        updated_critic_1 = matd3.critics_1[agent_id]
        assert_not_equal_state_dict(old_critic_1_sd, updated_critic_1.state_dict())

    for agent_id, old_critic_2_sd in critics_2_pre_learn_sd.items():
        updated_critic_2 = matd3.critics_2[agent_id]
        assert_not_equal_state_dict(old_critic_2_sd, updated_critic_2.state_dict())

    for agent_id, old_actor_target in matd3.actor_targets.items():
        updated_actor_target = matd3.actor_targets[agent_id]
        assert old_actor_target == updated_actor_target

    for agent_id, old_critic_target_1 in matd3.critic_targets_1.items():
        updated_critic_target_1 = matd3.critic_targets_1[agent_id]
        assert old_critic_target_1 == updated_critic_target_1

    for agent_id, old_critic_target_2 in matd3.critic_targets_2.items():
        updated_critic_target_2 = matd3.critic_targets_2[agent_id]
        assert old_critic_target_2 == updated_critic_target_2


@pytest.mark.parametrize("compile_mode", [None])
def test_matd3_soft_update(device, compile_mode, ma_vector_space, ma_discrete_space):
    accelerator = None

    matd3 = MATD3(
        observation_spaces=ma_vector_space,
        action_spaces=ma_discrete_space,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        accelerator=accelerator,
        device=device,
        torch_compiler=compile_mode,
    )

    for agent_id in matd3.agent_ids:
        actor = matd3.actors[agent_id]
        actor_target = matd3.actor_targets[agent_id]
        critic_1 = matd3.critics_1[agent_id]
        critic_target_1 = matd3.critic_targets_1[agent_id]
        critic_2 = matd3.critics_2[agent_id]
        critic_target_2 = matd3.critic_targets_2[agent_id]

        # Check actors
        matd3.soft_update(actor, actor_target)
        eval_params = list(actor.parameters())
        target_params = list(actor_target.parameters())
        expected_params = [
            matd3.tau * eval_param + (1.0 - matd3.tau) * target_param
            for eval_param, target_param in zip(eval_params, target_params)
        ]
        assert all(
            torch.allclose(expected_param, target_param)
            for expected_param, target_param in zip(expected_params, target_params)
        )
        matd3.soft_update(critic_1, critic_target_1)
        eval_params = list(critic_1.parameters())
        target_params = list(critic_target_1.parameters())
        expected_params = [
            matd3.tau * eval_param + (1.0 - matd3.tau) * target_param
            for eval_param, target_param in zip(eval_params, target_params)
        ]

        assert all(
            torch.allclose(expected_param, target_param)
            for expected_param, target_param in zip(expected_params, target_params)
        )
        matd3.soft_update(critic_2, critic_target_2)
        eval_params = list(critic_2.parameters())
        target_params = list(critic_target_2.parameters())
        expected_params = [
            matd3.tau * eval_param + (1.0 - matd3.tau) * target_param
            for eval_param, target_param in zip(eval_params, target_params)
        ]

        assert all(
            torch.allclose(expected_param, target_param)
            for expected_param, target_param in zip(expected_params, target_params)
        )


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space", "ma_image_space"])
@pytest.mark.parametrize("sum_score", [True, False])
@pytest.mark.parametrize("compile_mode", [None])
@pytest.mark.parametrize("vectorized", [True, False])
def test_matd3_algorithm_test_loop(
    observation_spaces,
    ma_discrete_space,
    device,
    compile_mode,
    sum_score,
    vectorized,
    request,
):
    observation_spaces = request.getfixturevalue(observation_spaces)
    accelerator = None

    # Define environment and algorithm
    if vectorized:
        env = make_multi_agent_vect_envs(
            DummyMultiEnv,
            2,
            **dict(
                observation_spaces=observation_spaces[0],
                action_spaces=ma_discrete_space,
            ),
        )
    else:
        env = DummyMultiEnv(observation_spaces[0], ma_discrete_space)

    matd3 = MATD3(
        observation_spaces,
        ma_discrete_space,
        agent_ids=["agent_0", "agent_1", "other_agent_0"],
        accelerator=accelerator,
        device=device,
        torch_compiler=compile_mode,
    )
    mean_score = matd3.test(env, max_steps=10, sum_scores=sum_score)
    if sum_score:
        assert isinstance(mean_score, float)
    else:
        assert isinstance(mean_score, np.ndarray)
        assert len(mean_score) == 3


@pytest.mark.parametrize("observation_spaces", ["ma_vector_space"])
@pytest.mark.parametrize("compile_mode", [None, "default"])
@pytest.mark.parametrize("accelerator_flag", [False, True])
@pytest.mark.parametrize("wrap", [True, False])
def test_matd3_clone_returns_identical_agent(
    accelerator_flag, wrap, compile_mode, observation_spaces, ma_vector_space, request
):
    # Clones the agent and returns an identical copy.
    observation_spaces = request.getfixturevalue(observation_spaces)
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    expl_noise = 0.1
    index = 0
    batch_size = 64
    lr_actor = 0.001
    lr_critic = 0.01
    learn_step = 5
    gamma = 0.95
    tau = 0.01
    mut = None
    actor_networks = None
    critic_networks = None
    policy_freq = 2
    device = "cpu"
    accelerator = Accelerator(device_placement=False) if accelerator_flag else None
    matd3 = MATD3(
        observation_spaces,
        ma_vector_space,
        agent_ids,
        expl_noise=expl_noise,
        index=index,
        policy_freq=policy_freq,
        batch_size=batch_size,
        lr_actor=lr_actor,
        lr_critic=lr_critic,
        learn_step=learn_step,
        gamma=gamma,
        tau=tau,
        mut=mut,
        actor_networks=actor_networks,
        critic_networks=critic_networks,
        device=device,
        accelerator=accelerator,
        torch_compiler=compile_mode,
        wrap=wrap,
    )

    clone_agent = matd3.clone(wrap=wrap)

    assert isinstance(clone_agent, MATD3)
    assert clone_agent.observation_spaces == matd3.observation_spaces
    assert clone_agent.action_spaces == matd3.action_spaces
    assert clone_agent.n_agents == matd3.n_agents
    assert clone_agent.agent_ids == matd3.agent_ids
    assert clone_agent.index == matd3.index
    assert clone_agent.batch_size == matd3.batch_size
    assert clone_agent.lr_actor == matd3.lr_actor
    assert clone_agent.lr_critic == matd3.lr_critic
    assert clone_agent.learn_step == matd3.learn_step
    assert clone_agent.gamma == matd3.gamma
    assert clone_agent.tau == matd3.tau
    assert clone_agent.device == matd3.device
    assert clone_agent.accelerator == matd3.accelerator
    assert clone_agent.torch_compiler == matd3.torch_compiler

    for agent_id in clone_agent.agent_ids:
        assert torch.equal(clone_agent.expl_noise[agent_id], matd3.expl_noise[agent_id])

        clone_actor = clone_agent.actors[agent_id]
        actor = matd3.actors[agent_id]
        assert_state_dicts_equal(clone_actor.state_dict(), actor.state_dict())

        clone_actor_target = clone_agent.actor_targets[agent_id]
        actor_target = matd3.actor_targets[agent_id]
        assert_state_dicts_equal(
            clone_actor_target.state_dict(), actor_target.state_dict()
        )

        clone_critic_1 = clone_agent.critics_1[agent_id]
        critic_1 = matd3.critics_1[agent_id]
        assert_state_dicts_equal(clone_critic_1.state_dict(), critic_1.state_dict())

        clone_critic_target_1 = clone_agent.critic_targets_1[agent_id]
        critic_target_1 = matd3.critic_targets_1[agent_id]
        assert_state_dicts_equal(
            clone_critic_target_1.state_dict(), critic_target_1.state_dict()
        )

        clone_critic_2 = clone_agent.critics_2[agent_id]
        critic_2 = matd3.critics_2[agent_id]
        assert_state_dicts_equal(clone_critic_2.state_dict(), critic_2.state_dict())

        clone_critic_target_2 = clone_agent.critic_targets_2[agent_id]
        critic_target_2 = matd3.critic_targets_2[agent_id]
        assert_state_dicts_equal(
            clone_critic_target_2.state_dict(), critic_target_2.state_dict()
        )


@pytest.mark.parametrize("compile_mode", [None, "default"])
def test_clone_new_index(compile_mode, ma_vector_space):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]

    matd3 = MATD3(
        ma_vector_space,
        copy.deepcopy(ma_vector_space),
        agent_ids,
        torch_compiler=compile_mode,
    )
    clone_agent = matd3.clone(index=100)

    assert clone_agent.index == 100


@pytest.mark.parametrize("compile_mode", [None])
def test_clone_after_learning(compile_mode, ma_vector_space):
    agent_ids = ["agent_0", "agent_1", "other_agent_0"]
    batch_size = 8

    matd3 = MATD3(
        ma_vector_space,
        copy.deepcopy(ma_vector_space),
        agent_ids,
        batch_size=batch_size,
        torch_compiler=compile_mode,
    )

    states = {
        agent_id: torch.randn(batch_size, ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    actions = {
        agent_id: torch.randn(batch_size, ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    rewards = {agent_id: torch.randn(batch_size, 1) for agent_id in agent_ids}
    next_states = {
        agent_id: torch.randn(batch_size, ma_vector_space[idx].shape[0])
        for idx, agent_id in enumerate(agent_ids)
    }
    dones = {agent_id: torch.zeros(batch_size, 1) for agent_id in agent_ids}

    experiences = states, actions, rewards, next_states, dones
    matd3.learn(experiences)
    clone_agent = matd3.clone()
    assert isinstance(clone_agent, MATD3)
    assert clone_agent.observation_spaces == matd3.observation_spaces
    assert clone_agent.action_spaces == matd3.action_spaces
    assert clone_agent.n_agents == matd3.n_agents
    assert clone_agent.agent_ids == matd3.agent_ids
    assert clone_agent.index == matd3.index
    assert clone_agent.batch_size == matd3.batch_size
    assert clone_agent.lr_actor == matd3.lr_actor
    assert clone_agent.lr_critic == matd3.lr_critic
    assert clone_agent.learn_step == matd3.learn_step
    assert clone_agent.gamma == matd3.gamma
    assert clone_agent.tau == matd3.tau
    assert clone_agent.device == matd3.device
    assert clone_agent.accelerator == matd3.accelerator
    assert clone_agent.torch_compiler == compile_mode
    assert matd3.torch_compiler == compile_mode

    for agent_id in clone_agent.agent_ids:
        assert torch.equal(clone_agent.expl_noise[agent_id], matd3.expl_noise[agent_id])

        clone_actor = clone_agent.actors[agent_id]
        actor = matd3.actors[agent_id]
        assert_state_dicts_equal(clone_actor.state_dict(), actor.state_dict())

        clone_actor_target = clone_agent.actor_targets[agent_id]
        actor_target = matd3.actor_targets[agent_id]
        assert_state_dicts_equal(
            clone_actor_target.state_dict(), actor_target.state_dict()
        )

        clone_critic_1 = clone_agent.critics_1[agent_id]
        critic_1 = matd3.critics_1[agent_id]
        assert_state_dicts_equal(clone_critic_1.state_dict(), critic_1.state_dict())

        clone_critic_target_1 = clone_agent.critic_targets_1[agent_id]
        critic_target_1 = matd3.critic_targets_1[agent_id]
        assert_state_dicts_equal(
            clone_critic_target_1.state_dict(), critic_target_1.state_dict()
        )

        clone_critic_2 = clone_agent.critics_2[agent_id]
        critic_2 = matd3.critics_2[agent_id]
        assert_state_dicts_equal(clone_critic_2.state_dict(), critic_2.state_dict())

        clone_critic_target_2 = clone_agent.critic_targets_2[agent_id]
        critic_target_2 = matd3.critic_targets_2[agent_id]
        assert_state_dicts_equal(
            clone_critic_target_2.state_dict(), critic_target_2.state_dict()
        )
